// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnClass.C   Version 1.8
//   LastSCCS    : 1/30/92  17:27:39
//   LastEdit    : "Mon Jan 20 10:36:46 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnClass,GnPresentable,);
ImplGenericListWithIterator(GnClassList,GnClassListIterator);

// static class variables

class GnClassList * GnClass::class_list = 0;
int GnClass::initialized = 0;
GnClass *GnClass::root = 0;
GnClassOrderingListIterator GnClass::oit_1, GnClass::oit_2;
//GnLinkedList GnClass::vbases;

GnClass::GnClass(char *n, unsigned int sz,
		 char *dfn, int dfl, char *ifn, int ifl,
		 GnBaseClassCollector bcc, GnClassMemberCollector cmc)
{
    GnClass *root = 0;
    
    name = strdup(n);
    size = sz;
    mi = 0;
    bc_collector = bcc;
    cm_collector = cmc;
    instance_count = 0;
    deleted_instance_count = 0;
    decl_file_name = dfn;
    decl_file_line = dfl;
    impl_file_name = ifn;
    impl_file_line = ifl;
    base_classes  = new GnClassInfoList;
    sub_classes   = new GnClassInfoList;
    //class_members = new GnClassMemberList;
    ordering      = new GnClassOrderingList;
    if( class_list == 0 )
	class_list = new GnClassList();
    class_list->Append(this);
}

int GnClass::
StoreOn(ostream &ostr, char *addr)
{
    ostr << (char *)addr << endl;
    return ostr.fail() ? 0 : 1;
}

int GnClass::
ReadFrom(istream &istr, char *addr)
{
    unsigned int ptr;
    
    istr >> ptr;
    *(void **)addr = (void *)ptr;
    return istr.fail() ? 0 : 1;
}

void GnClass::
Initialize()
{
    REQUIRE( ! IsInitialized(), "Classes are not initialized" );

    GnClassListIterator *it;

    initialized = 1;
    root = GnObject::ClassOf();

    // Compute the inheritance information
    
    for( it = ClassListIterator(); ! it->Off(); it->Forth() ) {
	if( it->Item()->bc_collector )
	    (it->Item()->bc_collector)((GnObject *)0,
				       it->Item()->base_classes);
//	if( it->Item()->cm_collector )
//	    (it->Item()->cm_collector)((GnObject *)0,
//				       it->Item()->class_members);
	GnClassInfoListIterator bases(it->Item()->base_classes);
	for( bases.Start(); ! bases.Off(); bases.Forth() ) {
	    bases.Item()->TheClass()->sub_classes
		->Append(new GnClassInfo(it->Item(),
					 bases.Item()->IsVirtual()));
	}
    }

    // Compute MI flag for all classes (1 = derived by MI, 0 = only SI )
    
    for( it->Start(); ! it->Off(); it->Forth() ) {
	GnClassInfoList *cil = it->Item()->base_classes;
	while( cil->Length() == 1 )
	    cil = cil->First()->TheClass()->base_classes;
	it->Item()->mi = cil->Length() > 1 ? 1 : 0;
    }
    delete it;
    
    // Compute ordering information for fast subclass check
    
    unsigned int pre  = 0;
    unsigned int post = 0;

    root->OrderNumbering(&pre, &post);
}    
    
void GnClass::
OrderNumbering(unsigned int *pre, unsigned int *post)
{
    // Append new ordering object

    (*pre)++;
    GnClassOrdering *o = new GnClassOrdering(*pre);
    ordering->Append( o );

    // Iterate forward on subclasses
    
    GnClassInfoListIterator it(sub_classes);
    for( it.Start(); ! it.Off(); it.Forth() ) {
	it.Item()->TheClass()->OrderNumbering(pre, post);
    }

    (*post)++;
    o->post_value = *post;
}

int GnClass::
IsSubclass(GnClass *cl)
{
    if( IsSimpleType() || cl->IsSimpleType() )
	return 0;
    
    if( ! mi ) {
	if(ordering->First()->pre_value >= cl->ordering->First()->pre_value &&
	   ordering->First()->post_value <= cl->ordering->First()->post_value )
	    return 1;
	else
	    return 0;
    } else {
	oit_1.Attach(ordering);
	oit_2.Attach(cl->ordering);
	for( oit_1.Start(); ! oit_1.Off(); oit_1.Forth() )
	    for( oit_2.Start(); ! oit_2.Off(); oit_2.Forth() )
		if(oit_1.Item()->pre_value >= oit_2.Item()->pre_value &&
		   oit_1.Item()->post_value <= oit_2.Item()->post_value ) {
		    oit_1.Detach();
		    oit_2.Detach();
		    return 1;
		}
	oit_1.Detach();
	oit_2.Detach();
	return 0;
    }
}

void *GnClass::
AmbiguousCast(void *first, void *second, GnClass *c)
{
    if( first == second || second == 0 )
	return first;
    if( first == 0 )
	return second;
    cerr << "Fatal: Ambiguous downward cast attempted: ";
    cerr << Name() << " into " << c->Name() << endl;
    abort();
    return 0;
}

void GnClass::
EnsureClassEqual(const GnClass *c)
{
    if( this != c ) {
	cerr << "Fatal: Classes not equal: " 
             << Name() << " and " << c->Name() << endl;
        abort();
    }
}

// int GnClass::
// BaseStored(void *base)
// {
//     if( vbases.member(base) )
// 	return 1;
//     else
// 	vbases.Append(base);
//     return 0;
// }
// 
// void GnClass::
// ClearBases()
// {
//     vbases.clear();
// }
 
class GnClassListIterator *GnClass::
ClassListIterator()
{
    return new GnClassListIterator(class_list);
}

class GnClassInfoListIterator *GnClass::
BaseClassesIterator()
{
    return new GnClassInfoListIterator(base_classes);
}

class GnClassInfoListIterator *GnClass::
SubclassesIterator()
{
    return new GnClassInfoListIterator(sub_classes);
}

char *GnClass::
PrintOn(char *buffer, unsigned int length)
{
    REQUIRE( IsStringPresentable(), "Object has a string representation");
    REQUIRE( length > 0, "Positive length");

    ostrstream str(buffer, length, ios::out);

    str << Name() << '\0';
    buffer[length-1] = '\0';

    return buffer;
}
