// This may look like C code, but it is really -*- C++ -*-

//   Module      : PlayerView.C   Version 1.2
//   LastSCCS    : 2/26/92  16:38:01
//   LastEdit    : "Thu Jan 30 14:15:37 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

PlayerView::
PlayerView(SleuthDocument* doc, Player *p)
: SleuthView( doc )
{
    player = p;
}

void PlayerView ::
button_press(int , int , int x, int y)
{
    int f_color, f_shape, f_value;
    
    //cout << "You hit ";
    switch( WhatsUnder(x, y, f_shape, f_color, f_value) ) {
      case eTableNothing:
	break;
      case ePlayer:
      case eField:
      case eShapeRow:
      case eValueRow:
      case eColorColumn:
	(new EnterRuleFeedback(this, x, y))->submit();
	break;
    }
}

void PlayerView::
draw(int, int, int, Dimension, Dimension)
{
    DrawEmptyTable();
    DrawTable();
}

void PlayerView::
UpdateField(Field *field)
{
    REQUIRE(DrawingIsPossible(), "Drawing into X Window is possible");    
    
    FieldStatus status = field->GetState();
    XGCValues val;

    switch( status ) {
      case eNeutral:
      case eNotOwned:
	val.foreground = status == eNotOwned ? foreground : background;
	val.background = status == eNotOwned ? background : foreground;
	change_gc(&val, GCForeground | GCBackground);

	XFillRectangle(get_display(), x_window, gcontext,
		       InnerX(FieldX(field->GetShape(),field->GetValue())),
		       InnerY(FieldY(field->GetColor())),
		       InnerWidth(TableFieldWidth()),
		       InnerHeight(TableFieldHeight()));
	break;
      case eOwned:
	val.foreground = foreground;
	val.background = background;
	val.line_width = 0;
	val.line_style =  LineSolid;
	change_gc(&val,
		  GCForeground | GCBackground | GCLineWidth | GCLineStyle);
	XClearArea(get_display(), x_window,
		   InnerX(FieldX(field->GetShape(),field->GetValue())),
		   InnerY(FieldY(field->GetColor())),
		   InnerWidth(TableFieldWidth()),
		   InnerHeight(TableFieldHeight()),
		   False);
	XDrawLine(get_display(), x_window, gcontext,
		  InnerX(FieldX(field->GetShape(),field->GetValue())),
		  InnerY(FieldY(field->GetColor())),
		  InnerX(FieldX(field->GetShape(),field->GetValue()))
		  + InnerWidth(TableFieldWidth()) - 1,
		  InnerY(FieldY(field->GetColor()))
		  + InnerHeight(TableFieldHeight()) - 1);
	XDrawLine(get_display(), x_window, gcontext,
		  InnerX(FieldX(field->GetShape(),field->GetValue())),
		  InnerY(FieldY(field->GetColor()))
		  + InnerHeight(TableFieldHeight()) - 1,
		  InnerX(FieldX(field->GetShape(),field->GetValue()))
		  + InnerWidth(TableFieldWidth()) - 1,
		  InnerY(FieldY(field->GetColor())));
	break;
    }
}

void PlayerView::
DrawSelected(Boolean state)
{
    DrawPlayerName(state);
}

void PlayerView::
DrawTable()
{
    int i, j, k;

    DrawSelected(player->IsSelected());
    
    for( i = 0; i < 3; i++)
	for( j = 0; j < 4; j++)
	    for( k = 0; k < 3; k++)
		UpdateField(player->GetField(i, j, k));
}

void PlayerView::
DrawPlayerName(Boolean reverse)
{
    DrawFilledString(player->GetName(),
		     strlen(player->GetName()),
		     reverse ? background : foreground,
		     reverse ? foreground : background,
		     InnerX(InnerX(TableX())),
		     InnerY(InnerY(TableY())),
		     InnerWidth(InnerWidth(CardColorColumnWidth())),
		     InnerHeight(InnerHeight(2*TableFieldHeight())));
}











