;; keybinding for standard default sunterm keys
;; Copyright (C) 1987 Free Software Foundation, Inc.

;; This file is part of GNU Emacs.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the GNU Emacs General Public
;; License for full details.

;; Everyone is granted permission to copy, modify and redistribute
;; GNU Emacs, but only under the conditions described in the
;; GNU Emacs General Public License.   A copy of this license is
;; supposed to have been given to you along with GNU Emacs so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

;;  Jeff Peck, Sun Microsystems Inc  <peck@sun.com>

(defun ignore-key ()
  "interactive version of ignore"
  (interactive)
  (ignore))

(defun scroll-down-in-place (n)
  (interactive "p")
  (previous-line n)
  (scroll-down n))

(defun scroll-up-in-place (n)
  (interactive "p")
  (next-line n)
  (scroll-up n))

(defun kill-region-and-unmark (beg end)
  "Like kill-region, but pops the mark [which equals point, anyway.]"
  (interactive "r")
  (kill-region beg end)
  (setq this-command 'kill-region-and-unmark)
  (set-mark-command t))

(defun prev-complex-command ()
  "Select Previous-complex-command"
  (interactive)
  (if (zerop (minibuffer-depth))
      (repeat-complex-command 1)
    (previous-complex-command 1)))

(defun rerun-prev-command ()
  "Repeat Previous-complex-command."
  (interactive)
  (eval (nth 0 command-history)))

(defvar grep-arg nil "Default arg for RE-search")
(defun grep-arg ()
  (if (memq last-command '(research-forward research-backward)) grep-arg
    (let* ((command (car command-history))
	   (command-name (symbol-name (car command)))
	   (search-arg (car (cdr command)))
	   (search-command 
	    (and command-name (string-match "search" command-name)))
	   )
      (if (and search-command (stringp search-arg)) (setq grep-arg search-arg)
	(setq search-command this-command 
	      grep-arg (read-string "REsearch: " grep-arg)
	      this-command search-command)
	grep-arg))))

(defun research-forward ()
  "Repeat RE search forward."
  (interactive)
  (re-search-forward (grep-arg)))

(defun research-backward ()
  "Repeat RE search backward."
  (interactive)
  (re-search-backward (grep-arg)))

;;;
;;; handle sun's extra function keys
;;; this version for those who run with standard .ttyswrc and no emacstool
;;;
;;; sunview picks up expose and open on the way UP, 
;;; so we ignore them on the way down
;;;

(defvar sun-esc-bracket t
  "*If non-nil, rebind ESC [ as prefix for Sun function keys.")

;;;(setq sun-esc-bracket t)

(defvar sun-raw-map (make-sparse-keymap) "*Keymap for ESC-[ encoded keyboard")

(define-key sun-raw-map "213z" 'scroll-up-in-place)	; R6
(define-key sun-raw-map "210z" 'scroll-down-in-place)	; R3
(define-key sun-raw-map "209z" 'shrink-window)		; R2
(define-key sun-raw-map "212z" 'enlarge-window)	        ; R5
(define-key sun-raw-map "214z" 'beginning-of-buffer)	; R7
(define-key sun-raw-map "216z" 'scroll-down)		; R9
(define-key sun-raw-map "218z" 'recenter)               ; R11
(define-key sun-raw-map "220z" 'end-of-buffer)		; R13
(define-key sun-raw-map "222z" 'scroll-up)		; R15
(define-key sun-raw-map "192z" 'enlarge-window-horizontally)		; L1 (Stop)
(define-key sun-raw-map "193z" 'shrink-window-horizontally)		; L2 (Again)
(define-key sun-raw-map "194z" 'list-buffers)		; L3 (Props)
(define-key sun-raw-map "195z" 'undo)			; L4 (Undo)
(define-key sun-raw-map "196z" 'ignore-key)		; L5 (Front)
(define-key sun-raw-map "197z" 'ignore-key)	        ; L6 (Copy)
(define-key sun-raw-map "198z" 'ignore-key)		; L7 (Open)
(define-key sun-raw-map "199z" 'ignore-key)	        ; L8 (Paste)
(define-key sun-raw-map "200z" 'exchange-point-and-mark); L9 (Find)
(define-key sun-raw-map "201z" 'kill-region-and-unmark)	; L10 (Cut)
;(define-key sun-raw-map "226z" 'scroll-down-in-place)	; t3
;(define-key sun-raw-map "227z" 'scroll-up-in-place)	; t4
;(define-key sun-raw-map "229z" 'shrink-window)		; t6
;(define-key sun-raw-map "230z" 'enlarge-window)	; t7

;;; my specials
;(define-key sun-raw-map "226z" 'insert-file-header)     ; F3: BAP-file-header
;(define-key sun-raw-map "225z" 'insert-defun-doc)       ; F2: BAP-doc-string for defun, etc.
;(define-key sun-raw-map "224z" 'insert-defclass-doc)    ; F1: BAP-doc-string for defclass, etc.


(if sun-esc-bracket
    (progn
      (define-key esc-map "[" sun-raw-map)		; Install sun-raw-map
      (define-key esc-map "[A" 'previous-line )		; R8
      (define-key esc-map "[B" 'next-line)		; R14
      (define-key esc-map "[C" 'forward-char)		; R12
      (define-key esc-map "[D" 'backward-char)		; R10
      (define-key esc-map "[[" 'backward-paragraph)	; the original esc-[
      ))


;;; Since .emacs gets loaded before this file, a hook is supplied
;;; for you to put your own bindings in.

(defvar sun-raw-map-hooks nil
  "List of forms to evaluate after setting sun-raw-map.")

(let ((hooks sun-raw-map-hooks))
  (while hooks
    (eval (car hooks))
    (setq hooks (cdr hooks))
    ))
