/* @(#)GnDocument.h	1.33 2/26/92 */

#ifndef Gina_Document_H
#define Gina_Document_H

class GnViewObject;
class GnDocumentShell;
class GnHistoryScroller;
class GnApplication;
class GnPathname;
class GnMessageDialog;
class GnFileSelectionDialog;
class GpathSelectionDialog;
class GnMenuEntry;


class GnDocument : public GnObject {
    META_DEF_1(GnDocument,GnObject);
    friend class GnBasicDocumentShell;
    //friend class GrawDocumentShell;
    friend class GnDocumentShell;
    friend class GnDocumentShellWithScroller;
    friend class GnHistoryScroller;
    friend class GnApplication;
  public:
    GnDocument ();
  public:
    virtual GnView *create_view()
	{ return new GnView( this ); }

    // Accessors:
    char * get_doc_title (); // liefert den Namen des Dokuments
    GnViewList *get_views ()
	{ return views; }
    Display *get_display ();
    GnDocumentShell *get_main_shell ()
	{ return main_shell; }
    GnView *get_main_view ()
	{ return main_view; }
    GnTopLevelShellList *get_toplevel_shells()
	{ return(toplevel_shells); };
    char *get_name()
	{ return(name); };
    XGCValues& gc_values() { return values; }
    void set_gc_defaults();

    void set_views (GnView*);
    void set_main_shell (GnDocumentShell*);
    void set_main_view (GnView*);
    void set_modified ( Boolean mod )
	{ modified = mod; }

    // Methods:
    virtual void create_windows(int width=0, int height=0);
    virtual void destroy_windows ();
    virtual void wipe_out();
    
    void resize_windows (int width, int height);
    void redraw_views ();
    void minifinder ();
    Boolean save_as ();
    Boolean save ();
    Boolean revert ();
    DialogResult do_close ();
  public:
    // Methods for handling command objects and command history
    virtual void command_executed();
    void execute_command_executed();
    void register_command( GnCommand * );
    void undo( Boolean update_history_scroller =TRUE);
    void irreversible_change();
    void redo( Boolean update_history_scroller =TRUE );
    void replay_history();
    void display_history_scroller();
  public:
    static void put_string(ostream &str, const char *s,
			   Boolean add_newline = True);
    static char *get_string(istream &str);
  protected:
    // Methods for file IO
    Boolean write_to_file(int fd);
    Boolean read_from_file(int fd, Boolean create_windows =TRUE);
    virtual Boolean write_to_stream( ostream &out_file );
    virtual Boolean read_from_stream( istream &in_file );
    void rename( GnPathname *new_pathname );
  protected:
    // Saving of command history
    virtual Boolean SaveHistory();
    virtual GnCommand *MakeNewCommand(char *p_command_class_name);
    Boolean WriteHistory(ostream &out_file);
    Boolean RestoreHistory(istream &in_file);
  protected:
    // Methods for printing documents
    void postscript_print ();
    void bitmap_print ();
    void print_object ( ostream & );

    /* Die Callback-Empfangsroutinen fuer obige Kommandos, werden benoetigt,
       um 'virtual' auszunutzen. */
    void receive_minifinder_callback(caddr_t call_data);
    void receive_do_close_callback(caddr_t call_data);
    void receive_save_as_callback(caddr_t call_data);
    void receive_save_callback(caddr_t call_data);
    void receive_revert_callback(caddr_t call_data);
    void receive_postscript_print_callback(caddr_t call_data);
    void receive_bitmap_print_callback(caddr_t call_data);
    void receive_undo_callback(caddr_t call_data);
    void receive_redo_callback(caddr_t call_data);
    void receive_replay_history_callback(caddr_t call_data);
    void receive_display_history_scroller_callback(caddr_t call_data);

    // Only used by the empty application
    void receive_activate_callback(caddr_t call_data);
  protected:
    char * create_new_name ();
    char * name;
    int number;
    GnPathname *file_pathname;

    Boolean modified;

    GnView *main_view;
    GnDocumentShell *main_shell;
    GnHistoryScroller *history_scroller;
    XGCValues values;		/* default values for CG */

    GnCommandList undo_commands;
    GnCommandList redo_commands;
    int max_history;
    // Update undo/redo menu entries and history scroller after cmd. execution
    void command_executed_update();

    // Lists for all shells and views (???)
    GnTopLevelShellList *toplevel_shells;
    GnViewList *views;

    // Standard menu entries
    GnMenuEntry *undo_menu_entry;
    GnMenuEntry *redo_menu_entry;
    GnMenuEntry *save_menu_entry;
    GnMenuEntry *revert_menu_entry;
    GnMenuEntry *history_scroller_menu_entry;
    GnMenuEntry *replay_history_menu_entry;

    Boolean write_header_to_stream(ostream &out_file);
    Boolean read_header_from_stream(istream &in_file,
				    Dimension *, Dimension *);
    // Use for some IO checking
    time_t creation_time;
    
    //void create_checkpoint ( iostream *os );
    //void restore_checkpoint ( iostream *is );
};

DefGenericListWithIterator(GnDocumentList,GnDocument,GnDocumentListIterator);

#endif Gina_Document_H
