// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnString.h   Version 1.2
//   LastSCCS    : 1/14/92  15:16:46
//   LastEdit    : "Fri Nov 29 12:50:28 1991"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#if defined(SHORT_NAMES) || defined(VMS)
#define re_compile_pattern	recmppat
#define re_pattern_buffer	repatbuf
#define re_registers		reregs
#endif

#ifndef _String_h
#ifdef __GNUG__
#  pragma once
#  pragma interface
#else
# define volatile
# define signed
#endif
#define _String_h 1

//#include <stream.h>
#include <Gina/GnRegex.h>

struct GnStrRep                     // internal GnString representations
{
  unsigned short    len;         // string length 
  unsigned short    sz;          // allocated space
  char              s[1];        // the string starts here 
                                 // (at least 1 char for trailing null)
                                 // allocated & expanded via non-public fcts
};

// primitive ops on StrReps -- nearly all GnString fns go through these.

GnStrRep*     Salloc(GnStrRep*, const char*, int, int);
GnStrRep*     Scopy(GnStrRep*, GnStrRep*);
GnStrRep*     Sresize(GnStrRep*, int);
GnStrRep*     Scat(GnStrRep*, const char*, int, const char*, int);
GnStrRep*     Scat(GnStrRep*, const char*, int,const char*,int, const char*,int);
GnStrRep*     Sprepend(GnStrRep*, const char*, int);
GnStrRep*     Sreverse(GnStrRep*, GnStrRep*);
GnStrRep*     Supcase(GnStrRep*, GnStrRep*);
GnStrRep*     Sdowncase(GnStrRep*, GnStrRep*);
GnStrRep*     Scapitalize(GnStrRep*, GnStrRep*);

// These classes need to be defined in the order given

class GnString;
class GnSubString;

class GnSubString
{
  friend class      GnString;
protected:

  GnString&           S;        // The GnString I'm a substring of
  unsigned short    pos;      // starting position in S's rep
  unsigned short    len;      // length of substring

  void              assign(GnStrRep*, const char*, int = -1);
                    GnSubString(GnString& x, int p, int l);
                    GnSubString(const GnSubString& x);

public:

// Note there are no public constructors. SubStrings are always
// created via GnString operations

                   ~GnSubString();

  void              operator =  (const GnString&     y);
  void              operator =  (const GnSubString&  y);
  void              operator =  (const char* t);
  void              operator =  (char        c);

// return 1 if target appears anywhere in GnSubString; else 0

  int               contains(char        c) const;
  int               contains(const GnString&     y) const;
  int               contains(const GnSubString&  y) const;
  int               contains(const char* t) const;
  int               contains(const GnRegex&       r) const;

// return 1 if target matches entire GnSubString

  int               matches(const GnRegex&  r) const;

// IO 

  friend ostream&   operator<<(ostream& s, const GnSubString& x);

// status

  int               length() const;
  int               empty() const;
  const char*       chars() const;

  int               OK() const; 

};


class GnString
{
  friend class      GnSubString;

protected:
  GnStrRep*           rep;   // Strings are pointers to their representations

// some helper functions

  int               search(int, int, const char*, int = -1) const;
  int               search(int, int, char) const;
  int               match(int, int, int, const char*, int = -1) const;
  int               _gsub(const char*, int, const char* ,int);
  int               _gsub(const GnRegex&, const char*, int);
  GnSubString         _substr(int, int);

public:

// constructors & assignment

                    GnString();
                    GnString(const GnString& x);
                    GnString(const GnSubString&  x);
                    GnString(const char* t);
                    GnString(const char* t, int len);
                    GnString(char c);

                    ~GnString();

  void              operator =  (const GnString&     y);
  void              operator =  (const char* y);
  void              operator =  (char        c);
  void              operator =  (const GnSubString&  y);

// concatenation

  void              operator += (const GnString&     y); 
  void              operator += (const GnSubString&  y);
  void              operator += (const char* t);
  void              operator += (char        c);

  void              prepend(const GnString&     y); 
  void              prepend(const GnSubString&  y);
  void              prepend(const char* t);
  void              prepend(char        c);


// procedural versions:
// concatenate first 2 args, store result in last arg

  friend void     cat(const GnString&, const GnString&, GnString&);
  friend void     cat(const GnString&, const GnSubString&, GnString&);
  friend void     cat(const GnString&, const char*, GnString&);
  friend void     cat(const GnString&, char, GnString&);

  friend void     cat(const GnSubString&, const GnString&, GnString&);
  friend void     cat(const GnSubString&, const GnSubString&, GnString&);
  friend void     cat(const GnSubString&, const char*, GnString&);
  friend void     cat(const GnSubString&, char, GnString&);

  friend void     cat(const char*, const GnString&, GnString&);
  friend void     cat(const char*, const GnSubString&, GnString&);
  friend void     cat(const char*, const char*, GnString&);
  friend void     cat(const char*, char, GnString&);

// double concatenation, by request. (yes, there are too many versions, 
// but if one is supported, then the others should be too...)
// Concatenate first 3 args, store in last arg

  friend void     cat(const GnString&,const GnString&, const GnString&,GnString&);
  friend void     cat(const GnString&,const GnString&,const GnSubString&,GnString&);
  friend void     cat(const GnString&,const GnString&, const char*, GnString&);
  friend void     cat(const GnString&,const GnString&, char, GnString&);
  friend void     cat(const GnString&,const GnSubString&,const GnString&,GnString&);
  friend void     cat(const GnString&,const GnSubString&,const GnSubString&,GnString&);
  friend void     cat(const GnString&,const GnSubString&, const char*, GnString&);
  friend void     cat(const GnString&,const GnSubString&, char, GnString&);
  friend void     cat(const GnString&,const char*, const GnString&,    GnString&);
  friend void     cat(const GnString&,const char*, const GnSubString&, GnString&);
  friend void     cat(const GnString&,const char*, const char*, GnString&);
  friend void     cat(const GnString&,const char*, char, GnString&);

  friend void     cat(const char*, const GnString&, const GnString&,GnString&);
  friend void     cat(const char*,const GnString&,const GnSubString&,GnString&);
  friend void     cat(const char*,const GnString&, const char*, GnString&);
  friend void     cat(const char*,const GnString&, char, GnString&);
  friend void     cat(const char*,const GnSubString&,const GnString&,GnString&);
  friend void     cat(const char*,const GnSubString&,const GnSubString&,GnString&);
  friend void     cat(const char*,const GnSubString&, const char*, GnString&);
  friend void     cat(const char*,const GnSubString&, char, GnString&);
  friend void     cat(const char*,const char*, const GnString&,    GnString&);
  friend void     cat(const char*,const char*, const GnSubString&, GnString&);
  friend void     cat(const char*,const char*, const char*, GnString&);
  friend void     cat(const char*,const char*, char, GnString&);


// searching & matching

// return position of target in string or -1 for failure

  int               index(char        c, int startpos = 0) const;      
  int               index(const GnString&     y, int startpos = 0) const;      
  int               index(const GnSubString&  y, int startpos = 0) const;      
  int               index(const char* t, int startpos = 0) const;  
  int               index(const GnRegex&      r, int startpos = 0) const;       

// return 1 if target appears anyhere in GnString; else 0

  int               contains(char        c) const;
  int               contains(const GnString&     y) const;
  int               contains(const GnSubString&  y) const;
  int               contains(const char* t) const;
  int               contains(const GnRegex&      r) const;

// return 1 if target appears anywhere after position pos 
// (or before, if pos is negative) in GnString; else 0

  int               contains(char        c, int pos) const;
  int               contains(const GnString&     y, int pos) const;
  int               contains(const GnSubString&  y, int pos) const;
  int               contains(const char* t, int pos) const;
  int               contains(const GnRegex&      r, int pos) const;

// return 1 if target appears at position pos in GnString; else 0

  int               matches(char        c, int pos = 0) const;
  int               matches(const GnString&     y, int pos = 0) const;
  int               matches(const GnSubString&  y, int pos = 0) const;
  int               matches(const char* t, int pos = 0) const;
  int               matches(const GnRegex&      r, int pos = 0) const;

//  return number of occurences of target in GnString

  int               freq(char        c) const; 
  int               freq(const GnString&     y) const;
  int               freq(const GnSubString&  y) const;
  int               freq(const char* t) const;

// GnSubString extraction

// Note that you can't take a substring of a const GnString, since
// this leaves open the possiblility of indirectly modifying the
// GnString through the GnSubString

  GnSubString         at(int         pos, int len);
  GnSubString         operator () (int         pos, int len); // synonym for at

  GnSubString         at(const GnString&     x, int startpos = 0); 
  GnSubString         at(const GnSubString&  x, int startpos = 0); 
  GnSubString         at(const char* t, int startpos = 0);
  GnSubString         at(char        c, int startpos = 0);
  GnSubString         at(const GnRegex&      r, int startpos = 0); 

  GnSubString         before(int          pos);
  GnSubString         before(const GnString&      x, int startpos = 0);
  GnSubString         before(const GnSubString&   x, int startpos = 0);
  GnSubString         before(const char*  t, int startpos = 0);
  GnSubString         before(char         c, int startpos = 0);
  GnSubString         before(const GnRegex&       r, int startpos = 0);

  GnSubString         through(int          pos);
  GnSubString         through(const GnString&      x, int startpos = 0);
  GnSubString         through(const GnSubString&   x, int startpos = 0);
  GnSubString         through(const char*  t, int startpos = 0);
  GnSubString         through(char         c, int startpos = 0);
  GnSubString         through(const GnRegex&       r, int startpos = 0);

  GnSubString         from(int          pos);
  GnSubString         from(const GnString&      x, int startpos = 0);
  GnSubString         from(const GnSubString&   x, int startpos = 0);
  GnSubString         from(const char*  t, int startpos = 0);
  GnSubString         from(char         c, int startpos = 0);
  GnSubString         from(const GnRegex&       r, int startpos = 0);

  GnSubString         after(int         pos);
  GnSubString         after(const GnString&     x, int startpos = 0);
  GnSubString         after(const GnSubString&  x, int startpos = 0);
  GnSubString         after(const char* t, int startpos = 0);
  GnSubString         after(char        c, int startpos = 0);
  GnSubString         after(const GnRegex&      r, int startpos = 0);


// deletion

// delete len chars starting at pos
  void              del(int         pos, int len);

// delete the first occurrence of target after startpos

  void              del(const GnString&     y, int startpos = 0);
  void              del(const GnSubString&  y, int startpos = 0);
  void              del(const char* t, int startpos = 0);
  void              del(char        c, int startpos = 0);
  void              del(const GnRegex&      r, int startpos = 0);

// global substitution: substitute all occurrences of pat with repl

  int               gsub(const GnString&     pat, const GnString&     repl);
  int               gsub(const GnSubString&  pat, const GnString&     repl);
  int               gsub(const char* pat, const GnString&     repl);
  int               gsub(const char* pat, const char* repl);
  int               gsub(const GnRegex&      pat, const GnString&     repl);

// friends & utilities

// split string into array res at separators; return number of elements

  friend int        split(const GnString& x, GnString res[], int maxn, 
                          const GnString& sep);
  friend int        split(const GnString& x, GnString res[], int maxn, 
                          const GnRegex&  sep);

  friend GnString     common_prefix(const GnString& x, const GnString& y, 
                                  int startpos = 0);
  friend GnString     common_suffix(const GnString& x, const GnString& y, 
                                  int startpos = -1);
  friend GnString     replicate(char        c, int n);
  friend GnString     replicate(const GnString&     y, int n);
  friend GnString     join(GnString src[], int n, const GnString& sep);

// simple builtin transformations

  friend GnString     reverse(const GnString& x);
  friend GnString     upcase(const GnString& x);
  friend GnString     downcase(const GnString& x);
  friend GnString     capitalize(const GnString& x);

// in-place versions of above

  void              reverse();
  void              upcase();
  void              downcase();
  void              capitalize();

// element extraction

  char&             operator [] (int i);
  char              elem(int i) const;
  char              firstchar() const;
  char              lastchar() const;

// conversion

                    operator const char*() const;
  const char*       chars() const;


// IO

  friend ostream&   operator<<(ostream& s, const GnString& x);
  friend ostream&   operator<<(ostream& s, const GnSubString& x);
  friend istream&   operator>>(istream& s, GnString& x);

  friend int        readline(istream& s, GnString& x, 
                             char terminator = '\n',
                             int discard_terminator = 1);

// status

  int               length() const;
  int               empty() const;

// preallocate some space for GnString
  void              alloc(int newsize);

// report current allocation (not length!)

  int               allocation() const;


  volatile void     error(const char* msg) const;

  int               OK() const;
};

typedef GnString StrTmp; // for backward compatibility

// other externs

int        compare(const GnString&    x, const GnString&     y);
int        compare(const GnString&    x, const GnSubString&  y);
int        compare(const GnString&    x, const char* y);
int        compare(const GnSubString& x, const GnString&     y);
int        compare(const GnSubString& x, const GnSubString&  y);
int        compare(const GnSubString& x, const char* y);
int        fcompare(const GnString&   x, const GnString&     y); // ignore case

extern GnStrRep  _nilStrRep;
extern GnString _nilString;

// other inlines

GnString operator + (const GnString& x, const GnString& y);
GnString operator + (const GnString& x, const GnSubString& y);
GnString operator + (const GnString& x, const char* y);
GnString operator + (const GnString& x, char y);
GnString operator + (const GnSubString& x, const GnString& y);
GnString operator + (const GnSubString& x, const GnSubString& y);
GnString operator + (const GnSubString& x, const char* y);
GnString operator + (const GnSubString& x, char y);
GnString operator + (const char* x, const GnString& y);
GnString operator + (const char* x, const GnSubString& y);

int operator==(const GnString& x, const GnString& y); 
int operator!=(const GnString& x, const GnString& y);
int operator> (const GnString& x, const GnString& y);
int operator>=(const GnString& x, const GnString& y);
int operator< (const GnString& x, const GnString& y);
int operator<=(const GnString& x, const GnString& y);
int operator==(const GnString& x, const GnSubString&  y);
int operator!=(const GnString& x, const GnSubString&  y);
int operator> (const GnString& x, const GnSubString&  y);
int operator>=(const GnString& x, const GnSubString&  y);
int operator< (const GnString& x, const GnSubString&  y);
int operator<=(const GnString& x, const GnSubString&  y);
int operator==(const GnString& x, const char* t);
int operator!=(const GnString& x, const char* t);
int operator> (const GnString& x, const char* t);
int operator>=(const GnString& x, const char* t);
int operator< (const GnString& x, const char* t);
int operator<=(const GnString& x, const char* t);
int operator==(const GnSubString& x, const GnString& y);
int operator!=(const GnSubString& x, const GnString& y);
int operator> (const GnSubString& x, const GnString& y);
int operator>=(const GnSubString& x, const GnString& y);
int operator< (const GnSubString& x, const GnString& y);
int operator<=(const GnSubString& x, const GnString& y);
int operator==(const GnSubString& x, const GnSubString&  y);
int operator!=(const GnSubString& x, const GnSubString&  y);
int operator> (const GnSubString& x, const GnSubString&  y);
int operator>=(const GnSubString& x, const GnSubString&  y);
int operator< (const GnSubString& x, const GnSubString&  y);
int operator<=(const GnSubString& x, const GnSubString&  y);
int operator==(const GnSubString& x, const char* t);
int operator!=(const GnSubString& x, const char* t);
int operator> (const GnSubString& x, const char* t);
int operator>=(const GnSubString& x, const char* t);
int operator< (const GnSubString& x, const char* t);
int operator<=(const GnSubString& x, const char* t);

//#if defined(__OPTIMIZE__) || defined(USE_LIBGXX_INLINES)


// status reports, needed before defining other things

inline int         GnString::length() const {  return rep->len; }
inline int         GnString::empty() const { return rep->len == 0; }
inline const char* GnString::chars() const { return &(rep->s[0]); }
inline int         GnString::allocation() const { return rep->sz; }
inline void        GnString::alloc(int newsize) { rep = Sresize(rep, newsize); }

inline int         GnSubString::length() const { return len; }
inline int         GnSubString::empty() const { return len == 0; }
inline const char* GnSubString::chars() const { return &(S.rep->s[pos]); }


// constructors

inline GnString::GnString() 
  : rep(&_nilStrRep) {}
inline GnString::GnString(const GnString& x) 
  : rep(Scopy(0, x.rep)) {}
inline GnString::GnString(const char* t) 
  : rep(Salloc(0, t, -1, -1)) {}
inline GnString::GnString(const char* t, int tlen)
  : rep(Salloc(0, t, tlen, tlen)) {}
inline GnString::GnString(const GnSubString& y)
  : rep(Salloc(0, y.chars(), y.length(), y.length())) {}
inline GnString::GnString(char c) 
  : rep(Salloc(0, &c, 1, 1)) {}

inline GnString::~GnString() { if (rep != &_nilStrRep) delete rep; }

inline GnSubString::GnSubString(const GnSubString& x)
  :S(x.S), pos(x.pos), len(x.len) {}
inline GnSubString::GnSubString(GnString& x, int first, int l)
  :S(x), pos(first), len(l) {}

inline GnSubString::~GnSubString() {}

// assignment

inline void GnString::operator =  (const GnString& y)
{ 
  rep = Scopy(rep, y.rep);
}

inline void GnString::operator=(const char* t)
{
  rep = Salloc(rep, t, -1, -1); 
}

inline void GnString::operator=(const GnSubString&  y)
{
  rep = Salloc(rep, y.chars(), y.length(), y.length());
}

inline void GnString::operator=(char c)
{
  rep = Salloc(rep, &c, 1, 1); 
}


inline void GnSubString::operator = (const char* ys)
{
  assign(0, ys);
}

inline void GnSubString::operator = (char ch)
{
  assign(0, &ch, 1);
}

inline void GnSubString::operator = (const GnString& y)
{
  assign(y.rep, y.chars(), y.length());
}

inline void GnSubString::operator = (const GnSubString& y)
{
  assign(y.S.rep, y.chars(), y.length());
}

// Zillions of cats...

inline void cat(const GnString& x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnString& x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnString& x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y, -1);
}

inline void cat(const GnString& x, char y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), &y, 1);
}

inline void cat(const GnSubString& x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnSubString& x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnSubString& x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), y, -1);
}

inline void cat(const GnSubString& x, char y, GnString& r)
{
  r.rep = Scat(r.rep, x.chars(), x.length(), &y, 1);
}

inline void cat(const char* x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, x, -1, y.chars(), y.length());
}

inline void cat(const char* x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, x, -1, y.chars(), y.length());
}

inline void cat(const char* x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, x, -1, y, -1);
}

inline void cat(const char* x, char y, GnString& r)
{
  r.rep = Scat(r.rep, x, -1, &y, 1);
}

inline void cat(const GnString& a, const GnString& x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnString& a, const GnString& x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnString& a, const GnString& x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y, -1);
}

inline void cat(const GnString& a, const GnString& x, char y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), &y, 1);
}

inline void cat(const GnString& a, const GnSubString& x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnString& a, const GnSubString& x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const GnString& a, const GnSubString& x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), y, -1);
}

inline void cat(const GnString& a, const GnSubString& x, char y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x.chars(), x.length(), &y, 1);
}

inline void cat(const GnString& a, const char* x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, y.chars(), y.length());
}

inline void cat(const GnString& a, const char* x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, y.chars(), y.length());
}

inline void cat(const GnString& a, const char* x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, y, -1);
}

inline void cat(const GnString& a, const char* x, char y, GnString& r)
{
  r.rep = Scat(r.rep, a.chars(), a.length(), x, -1, &y, 1);
}


inline void cat(const char* a, const GnString& x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const GnString& x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const GnString& x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y, -1);
}

inline void cat(const char* a, const GnString& x, char y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), &y, 1);
}

inline void cat(const char* a, const GnSubString& x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const GnSubString& x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y.chars(), y.length());
}

inline void cat(const char* a, const GnSubString& x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), y, -1);
}

inline void cat(const char* a, const GnSubString& x, char y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x.chars(), x.length(), &y, 1);
}

inline void cat(const char* a, const char* x, const GnString& y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y.chars(), y.length());
}

inline void cat(const char* a, const char* x, const GnSubString& y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y.chars(), y.length());
}

inline void cat(const char* a, const char* x, const char* y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, y, -1);
}

inline void cat(const char* a, const char* x, char y, GnString& r)
{
  r.rep = Scat(r.rep, a, -1, x, -1, &y, 1);
}


// operator versions

inline void GnString::operator +=(const GnString& y)
{
  cat(*this, y, *this);
}

inline void GnString::operator +=(const GnSubString& y)
{
  cat(*this, y, *this);
}

inline void GnString::operator += (const char* y)
{
  cat(*this, y, *this);
}

inline void GnString:: operator +=(char y)
{
  cat(*this, y, *this);
}

// constructive concatenation

#if defined(__GNUG__) && !defined(NO_NRV)

inline GnString operator + (const GnString& x, const GnString& y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnString& x, const GnSubString& y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnString& x, const char* y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnString& x, char y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnSubString& x, const GnString& y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnSubString& x, const GnSubString& y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnSubString& x, const char* y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const GnSubString& x, char y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const char* x, const GnString& y) return r;
{
  cat(x, y, r);
}

inline GnString operator + (const char* x, const GnSubString& y) return r;
{
  cat(x, y, r);
}

inline GnString reverse(const GnString& x) return r;
{
  r.rep = Sreverse(x.rep, r.rep);
}

inline GnString upcase(const GnString& x) return r;
{
  r.rep = Supcase(x.rep, r.rep);
}

inline GnString downcase(const GnString& x) return r;
{
  r.rep = Sdowncase(x.rep, r.rep);
}

inline GnString capitalize(const GnString& x) return r;
{
  r.rep = Scapitalize(x.rep, r.rep);
}

#else /* NO_NRV */

inline GnString operator + (const GnString& x, const GnString& y)
{
  GnString r;  cat(x, y, r);  return r;
}

inline GnString operator + (const GnString& x, const GnSubString& y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const GnString& x, const char* y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const GnString& x, char y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const GnSubString& x, const GnString& y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const GnSubString& x, const GnSubString& y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const GnSubString& x, const char* y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const GnSubString& x, char y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const char* x, const GnString& y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString operator + (const char* x, const GnSubString& y) 
{
  GnString r; cat(x, y, r); return r;
}

inline GnString reverse(const GnString& x) 
{
  GnString r; r.rep = Sreverse(x.rep, r.rep); return r;
}

inline GnString upcase(const GnString& x) 
{
  GnString r; r.rep = Supcase(x.rep, r.rep); return r;
}

inline GnString downcase(const GnString& x) 
{
  GnString r; r.rep = Sdowncase(x.rep, r.rep); return r;
}

inline GnString capitalize(const GnString& x) 
{
  GnString r; r.rep = Scapitalize(x.rep, r.rep); return r;
}

#endif

// prepend

inline void GnString::prepend(const GnString& y)
{
  rep = Sprepend(rep, y.chars(), y.length());
}

inline void GnString::prepend(const char* y)
{
  rep = Sprepend(rep, y, -1); 
}

inline void GnString::prepend(char y)
{
  rep = Sprepend(rep, &y, 1); 
}

inline void GnString::prepend(const GnSubString& y)
{
  rep = Sprepend(rep, y.chars(), y.length());
}

// misc transformations


inline void GnString::reverse()
{
  rep = Sreverse(rep, rep);
}


inline void GnString::upcase()
{
  rep = Supcase(rep, rep);
}


inline void GnString::downcase()
{
  rep = Sdowncase(rep, rep);
}


inline void GnString::capitalize()
{
  rep = Scapitalize(rep, rep);
}

// element extraction

inline char&  GnString::operator [] (int i) 
{ 
  if (((unsigned)i) >= length()) error("invalid index");
  return rep->s[i];
}

inline char  GnString::elem (int i) const
{ 
  if (((unsigned)i) >= length()) error("invalid index");
  return rep->s[i];
}

inline char  GnString::firstchar() const
{ 
  return elem(0);
}

inline char  GnString::lastchar() const
{ 
  return elem(length() - 1);
}

// searching

inline int GnString::index(char c, int startpos) const
{
  return search(startpos, length(), c);
}

inline int GnString::index(const char* t, int startpos) const
{   
  return search(startpos, length(), t);
}

inline int GnString::index(const GnString& y, int startpos) const
{   
  return search(startpos, length(), y.chars(), y.length());
}

inline int GnString::index(const GnSubString& y, int startpos) const
{   
  return search(startpos, length(), y.chars(), y.length());
}

inline int GnString::index(const GnRegex& r, int startpos) const
{
  int unused;  return r.search(chars(), length(), unused, startpos);
}

inline int GnString::contains(char c) const
{
  return search(0, length(), c) >= 0;
}

inline int GnString::contains(const char* t) const
{   
  return search(0, length(), t) >= 0;
}

inline int GnString::contains(const GnString& y) const
{   
  return search(0, length(), y.chars(), y.length()) >= 0;
}

inline int GnString::contains(const GnSubString& y) const
{   
  return search(0, length(), y.chars(), y.length()) >= 0;
}

inline int GnString::contains(char c, int p) const
{
  return match(p, length(), 0, &c, 1) >= 0;
}

inline int GnString::contains(const char* t, int p) const
{
  return match(p, length(), 0, t) >= 0;
}

inline int GnString::contains(const GnString& y, int p) const
{
  return match(p, length(), 0, y.chars(), y.length()) >= 0;
}

inline int GnString::contains(const GnSubString& y, int p) const
{
  return match(p, length(), 0, y.chars(), y.length()) >= 0;
}

inline int GnString::contains(const GnRegex& r) const
{
  int unused;  return r.search(chars(), length(), unused, 0) >= 0;
}

inline int GnString::contains(const GnRegex& r, int p) const
{
  return r.match(chars(), length(), p) >= 0;
}


inline int GnString::matches(const GnSubString& y, int p) const
{
  return match(p, length(), 1, y.chars(), y.length()) >= 0;
}

inline int GnString::matches(const GnString& y, int p) const
{
  return match(p, length(), 1, y.chars(), y.length()) >= 0;
}

inline int GnString::matches(const char* t, int p) const
{
  return match(p, length(), 1, t) >= 0;
}

inline int GnString::matches(char c, int p) const
{
  return match(p, length(), 1, &c, 1) >= 0;
}

inline int GnString::matches(const GnRegex& r, int p) const
{
  int l = (p < 0)? -p : length() - p;
  return r.match(chars(), length(), p) == l;
}


inline int GnSubString::contains(const char* t) const
{   
  return S.search(pos, pos+len, t) >= 0;
}

inline int GnSubString::contains(const GnString& y) const
{   
  return S.search(pos, pos+len, y.chars(), y.length()) >= 0;
}

inline int GnSubString::contains(const GnSubString&  y) const
{   
  return S.search(pos, pos+len, y.chars(), y.length()) >= 0;
}

inline int GnSubString::contains(char c) const
{
  return S.search(pos, pos+len, 0, c) >= 0;
}

inline int GnSubString::contains(const GnRegex& r) const
{
  int unused;  return r.search(chars(), len, unused, 0) >= 0;
}

inline int GnSubString::matches(const GnRegex& r) const
{
  return r.match(chars(), len, 0) == len;
}


inline int GnString::gsub(const GnString& pat, const GnString& r)
{
  return _gsub(pat.chars(), pat.length(), r.chars(), r.length());
}

inline int GnString::gsub(const GnSubString&  pat, const GnString& r)
{
  return _gsub(pat.chars(), pat.length(), r.chars(), r.length());
}

inline int GnString::gsub(const GnRegex& pat, const GnString& r)
{
  return _gsub(pat, r.chars(), r.length());
}

inline int GnString::gsub(const char* pat, const GnString& r)
{
  return _gsub(pat, -1, r.chars(), r.length());
}

inline int GnString::gsub(const char* pat, const char* r)
{
  return _gsub(pat, -1, r, -1);
}


inline GnString::operator const char*() const
{ 
  return str(chars());
}

inline  ostream& operator<<(ostream& s, const GnString& x)
{
#ifdef VMS
   s << x.chars(); return s;
#else
#ifdef __GNUG__
  s.put(x.chars()); return s;
#else
   s << x.chars(); return s;
#endif
#endif

}

// a zillion comparison operators

inline int operator==(const GnString& x, const GnString& y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const GnString& x, const GnString& y)
{
  return compare(x, y) != 0; 
}

inline int operator>(const GnString& x, const GnString& y)
{
  return compare(x, y) > 0; 
}

inline int operator>=(const GnString& x, const GnString& y)
{
  return compare(x, y) >= 0; 
}

inline int operator<(const GnString& x, const GnString& y)
{
  return compare(x, y) < 0; 
}

inline int operator<=(const GnString& x, const GnString& y)
{
  return compare(x, y) <= 0; 
}

inline int operator==(const GnString& x, const GnSubString&  y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const GnString& x, const GnSubString&  y)
{
  return compare(x, y) != 0; 
}

inline int operator>(const GnString& x, const GnSubString&  y)      
{
  return compare(x, y) > 0; 
}

inline int operator>=(const GnString& x, const GnSubString&  y)
{
  return compare(x, y) >= 0; 
}

inline int operator<(const GnString& x, const GnSubString&  y) 
{
  return compare(x, y) < 0; 
}

inline int operator<=(const GnString& x, const GnSubString&  y)
{
  return compare(x, y) <= 0; 
}

inline int operator==(const GnString& x, const char* t) 
{
  return compare(x, t) == 0; 
}

inline int operator!=(const GnString& x, const char* t) 
{
  return compare(x, t) != 0; 
}

inline int operator>(const GnString& x, const char* t)  
{
  return compare(x, t) > 0; 
}

inline int operator>=(const GnString& x, const char* t) 
{
  return compare(x, t) >= 0; 
}

inline int operator<(const GnString& x, const char* t)  
{
  return compare(x, t) < 0; 
}

inline int operator<=(const GnString& x, const char* t) 
{
  return compare(x, t) <= 0; 
}

inline int operator==(const GnSubString& x, const GnString& y) 
{
  return compare(y, x) == 0; 
}

inline int operator!=(const GnSubString& x, const GnString& y)
{
  return compare(y, x) != 0;
}

inline int operator>(const GnSubString& x, const GnString& y)      
{
  return compare(y, x) < 0;
}

inline int operator>=(const GnSubString& x, const GnString& y)     
{
  return compare(y, x) <= 0;
}

inline int operator<(const GnSubString& x, const GnString& y)      
{
  return compare(y, x) > 0;
}

inline int operator<=(const GnSubString& x, const GnString& y)     
{
  return compare(y, x) >= 0;
}

inline int operator==(const GnSubString& x, const GnSubString&  y) 
{
  return compare(x, y) == 0; 
}

inline int operator!=(const GnSubString& x, const GnSubString&  y)
{
  return compare(x, y) != 0;
}

inline int operator>(const GnSubString& x, const GnSubString&  y)      
{
  return compare(x, y) > 0;
}

inline int operator>=(const GnSubString& x, const GnSubString&  y)
{
  return compare(x, y) >= 0;
}

inline int operator<(const GnSubString& x, const GnSubString&  y) 
{
  return compare(x, y) < 0;
}

inline int operator<=(const GnSubString& x, const GnSubString&  y)
{
  return compare(x, y) <= 0;
}

inline int operator==(const GnSubString& x, const char* t) 
{
  return compare(x, t) == 0; 
}

inline int operator!=(const GnSubString& x, const char* t) 
{
  return compare(x, t) != 0;
}

inline int operator>(const GnSubString& x, const char* t)  
{
  return compare(x, t) > 0; 
}

inline int operator>=(const GnSubString& x, const char* t) 
{
  return compare(x, t) >= 0; 
}

inline int operator<(const GnSubString& x, const char* t)  
{
  return compare(x, t) < 0; 
}

inline int operator<=(const GnSubString& x, const char* t) 
{
  return compare(x, t) <= 0; 
}


// a helper needed by at, before, etc.

inline GnSubString GnString::_substr(int first, int l)
{
  if (first < 0 || (unsigned)(first + l) > length()) 
    return GnSubString(_nilString, 0, 0) ;
  else 
    return GnSubString(*this, first, l);
}


//#endif

#endif
