// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnDocument.C   Version 1.43
//   LastSCCS    : 2/26/92  16:36:15
//   LastEdit    : "Tue Feb 25 09:29:30 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnDocument,GnObject,);

ImplGenericListWithIterator(GnDocumentList,GnDocumentListIterator);

GnDocument ::
GnDocument ()
{
    modified = FALSE;
    main_view = 0;
    main_shell = 0;
    history_scroller = 0;
    max_history = 0;
    set_gc_defaults();

    toplevel_shells = new GnTopLevelShellList;
    views = new GnViewList;

    number = GnApplication::get()->new_document_number();
    file_pathname = 0;
    name = 0;
    rename(0);
    creation_time = time((time_t *)0);
    Gina_Debug("creation_time"); Gina_Debug_NL(creation_time);
}

Display * GnDocument ::
get_display ()
{
    Require( GnApplication::get() != 0 );

    return GnApplication::get()->get_display();
}

void GnDocument ::
set_gc_defaults()
{
    values.function = GXcopy;
    values.plane_mask = ~0L;
    values.foreground = 1;
    values.background = 0;
    values.line_width = 0;
    values.line_style = LineSolid;
    values.cap_style = CapButt;
    values.join_style = JoinMiter;
    values.fill_style = FillSolid;
    values.fill_rule = EvenOddRule;
    values.arc_mode = ArcPieSlice;
    values.tile = 0;
    values.stipple = 0;
    values.ts_x_origin = 0;
    values.ts_y_origin = 0;
    values.font = 0;
    values.subwindow_mode = ClipByChildren;
    values.graphics_exposures = TRUE;
    values.clip_x_origin = 0;
    values.clip_y_origin = 0;
    values.clip_mask = None;
    values.dash_offset = 0;
    values.dashes = 4;
}

void GnDocument ::
set_main_shell (GnDocumentShell* shell)
{
    Require( shell != 0 );

    main_shell = shell;
}

void GnDocument ::
set_main_view ( GnView *view )
{
    Require( view != 0 );

    main_view = view;
}

char * GnDocument ::
create_new_name ()
{
    char *ret = new char[30];

    sprintf(ret, "%s%d", "Untitled-", number);
    return ret;
}

void GnDocument ::
print_object( ostream & )
{
}

void GnDocument ::
create_windows(int width, int height)
{
    main_shell = new GnDocumentShellWithScroller(this, width, height);
    main_shell->create(GnApplication::get()->get_application_shell(),
		       "mainShell");

    GnPushButton *pushButton[10];
    long d = 0x7fffffffL / 950L;
    srand(7);
    for ( int i = 0; i < 10; i++ ) {
	pushButton[i] = new GnPushButton;
	pushButton[i]->setR_x(short(rand()/d));
	pushButton[i]->setR_y(short(rand()/d));
	pushButton[i]->create( main_view, "Beep-Button" );
	pushButton[i]->add_activateCallback(CALLBACK(GnDocument,
						     receive_activate_callback,
						     this));
    }
//    main_view->set_input_callback();
}

void GnDocument ::
destroy_windows ()
{
    GnTopLevelShell *shell;

    wipe_out();

    while( ! toplevel_shells->Empty() ) {
	shell = toplevel_shells->RemoveFirst();
	shell->destroy();
	delete shell;
    }
    // Already deleted in while loop
    //main_shell->destroy();
    //delete main_shell;
    main_shell = 0;
}

void GnDocument::
wipe_out()
{
    // Wipe out the contents of a document
}

void GnDocument ::
rename( GnPathname *new_pathname )
{
    if( file_pathname && file_pathname != new_pathname)
	delete file_pathname;
    if( name )
	delete name;

    file_pathname = new_pathname;
    name = file_pathname ? strdup(file_pathname->name()) : create_new_name();
    if( main_shell ) {
	main_shell->setR_title(main_shell->
			       make_title(GnApplication::get()->app_name(),
					  name));
	main_shell->setR_iconName(name);
	main_shell->SetValues();
    }
}

void GnDocument ::
resize_windows(int new_width, int new_height)
{
    main_shell->resize(new_width, new_height);
}

void GnDocument ::
redraw_views()
{
    for(GnViewListIterator it(views); ! it.Off(); it.Forth() )
	it.Item()->force_redraw();
}

void GnDocument ::
minifinder ()
{
    GnPathname *out_path;

    if( select_old_file(*main_shell, &out_path, "", "", "", O_RDONLY,
			" Start file " ) == OK ) {
	Gina_Debug("Starting file: "); Gina_Debug_NL(out_path->abs_namestring());
	// start_file( path, this );
    }
}

Boolean GnDocument ::
save_as ()
{
    GnPathname *new_path;

    if( (file_pathname &&
	 select_old_or_new_file(*main_shell, &new_path, 
				file_pathname->name(),
				file_pathname->type(),
				file_pathname->directory(),
				O_WRONLY, " Save as ",
				creation_time) == OK ) ||
       (! file_pathname &&
	select_old_or_new_file(*main_shell, &new_path, name,
			       GnApplication::get()->app_file_type(), "",
			       O_WRONLY,  " Save as ", creation_time) == OK)) {
	rename(new_path);
	if( save_menu_entry && ! save_menu_entry->get_sensitive() )
	    save_menu_entry->set_sensitive(True);
	return(save());
    }

//    message_dialog(main_shell, "Save As ...");
//    error_dialog(main_shell, "Save As ...");
//    information_dialog(main_shell, "Save As ...");
//    question_dialog(main_shell, "Save As ...");
//    warning_dialog(main_shell, "Save As ...");
//    working_dialog(main_shell, "Save As ...");

    return(False);
}

Boolean GnDocument ::
save ()
{
    if( ! file_pathname ) {
	save_as();
    } else {
	int fd;
	
	if( (fd = file_pathname->open_fd(O_WRONLY | O_CREAT | O_TRUNC,
					 0777)) == -1 ) {
	    char *fmt = "Can't open output file \n\n  %s\n\n for writing!";
	    char *msg =
		new char[strlen(fmt)+strlen(file_pathname->abs_namestring())];
	    sprintf(msg, fmt, file_pathname->abs_namestring());
	    error_dialog(*main_shell, msg);
	    delete msg;
	    return(False);
	} else {
	    if( write_to_file( fd )) {
		set_modified(FALSE);
		return(True);
	    } else {
		error_dialog(*main_shell, "Write to file failed!");
		return(False);
	    }
	}
    }
}

Boolean GnDocument ::
revert ()
{
    int fd;

    if ( file_pathname && modified &&
	question_dialog(*main_shell,
			"Revert to last saved version of document ?") == Yes ) {
	wipe_out();
	if( (fd = file_pathname->open_fd(O_RDONLY)) == -1 ) {
	    error_dialog(*main_shell, "Can't open document file for reading!");
	    return(False);
	}
	if( read_from_file(fd, FALSE ) ) {
	    redraw_views();
	    irreversible_change();
	    return(True);
	} else {
	    return(False);
	}
    } else {
	return(False);
    }
}

DialogResult GnDocument ::
do_close ()
{
    DialogResult answer = OK;

    // Ask user only if document is modified
    if( modified) {
	char *fmt = "Save document\n\n   %s\n\nfirst?";
	char *fname = file_pathname ? file_pathname->abs_namestring() : name;
	char *msg = new char[strlen(fmt)+strlen(fname)];
	sprintf(msg, fmt, fname);
	answer = alternative_dialog(*main_shell, msg);
	delete msg;
	// answer may be OK, No, or Cancel
	if ( answer == OK )
	    // Return "Cancel" if an error occured during save()
	    if( ! save() )
		answer = Cancel;
    }
    // Destroy if answer is OK or No
    if ( answer != Cancel )
	destroy_windows();
    // map No to OK
    return(answer == Cancel ? Cancel : OK);
}

void GnDocument ::
execute_command_executed()
{
    command_executed();
    command_executed_update();
}

void GnDocument ::
command_executed() {}

void GnDocument ::
command_executed_update()
{
    Gina_Debug_NL("GnDocument :: command_executed_update");

    char *str;
    Boolean free_it = FALSE;
    if ( ! undo_commands.Empty() ) {
        str = new char [ strlen(undo_commands.First()->name())+6 ];
        strcpy( str, "Undo " );
        strcat( str, undo_commands.First()->name() );
        free_it = TRUE;
    }
    else
        str = "Undo";
    undo_menu_entry->set_label_string( str );
    undo_menu_entry->set_sensitive( ! undo_commands.Empty() );
    if ( free_it ) {
        free_it = FALSE;
        delete str;
    }

    if ( ! redo_commands.Empty() ) {
        str = new char [ strlen(redo_commands.First()->name())+6 ];
        strcpy( str, "Redo " );
        strcat( str, redo_commands.First()->name() );
        free_it = TRUE;
    }
    else
        str = "Redo";
    redo_menu_entry->set_label_string( str );
    redo_menu_entry->set_sensitive( ! redo_commands.Empty() );
    if ( free_it )
        delete str;

    if ( history_scroller )
        history_scroller->update();
}

void GnDocument ::
register_command( GnCommand *cmd )
{
    // Insert command
    undo_commands.Prepend( cmd );

    // Commit further redo
    while ( ! redo_commands.Empty() ) {
	redo_commands.RemoveFirst()->commit( TRUE );
    }

    // Adjust history
    if ( max_history && (undo_commands.Length() > max_history) ) {
	while( undo_commands.Length() > max_history ) {
	    undo_commands.RemoveLast()->commit( FALSE );
	}
    }
    execute_command_executed();
}

void GnDocument ::
undo ( Boolean /* upd_his_scr */ )
{
    Require( ! undo_commands.Empty() );
    
    GnCommand *cmd;

    cmd = undo_commands.RemoveFirst();
    cmd->execute_undoit();
    redo_commands.Prepend( cmd );
    if( cmd->causes_change() )
	set_modified(True);
    execute_command_executed();
}

void GnDocument ::
irreversible_change ()
{
    GnCommand *cmd;

    while( ! redo_commands.Empty() ) {
	cmd = redo_commands.RemoveFirst();
	cmd->commit( TRUE );
    }

    while( ! undo_commands.Empty() ) {
	cmd = undo_commands.RemoveFirst();
	cmd->commit( FALSE );
    }

    execute_command_executed();
}

void GnDocument ::
redo( Boolean )
{
    Require( ! redo_commands.Empty() );

    GnCommand *cmd;
    
    cmd = redo_commands.RemoveFirst();
    cmd->redoit();
    undo_commands.Prepend( cmd );
    if( cmd->causes_change() )
	set_modified(True);
    execute_command_executed();
}

void GnDocument ::
replay_history ()
{
    int nr_of_commands = undo_commands.Length();
    int i;

    if( history_scroller )
	history_scroller->Busy(True);
    for (i = 0; i < nr_of_commands; i++) {
	//sleep( 0.3 );
	undo();
	XFlush( get_display() );
    }
    for ( i=0; i < nr_of_commands; i++) {
	//sleep( 0.3 );
	redo();
	XFlush( get_display() );
    }
    if( history_scroller ) {
	history_scroller->Busy(False);
	history_scroller->update();
    }
}

void GnDocument::
put_string(ostream &ostr, const char *s, Boolean add_newline)
{
    ostr << strlen(s) << " " << s;
    if( add_newline )
	ostr << '\n';
    else
	ostr << ' ';
}

char *GnDocument::
get_string(istream &str)
{
    int length;
    char *string;
    char  whitespace;

    str >> length;
    whitespace = str.get();
    CHECK(whitespace == ' ', "Valid input format");
    string = new char[length+1];
    str.get(string, length+1);
    //cout << "Read<" << string <<">\n"; cout.flush();
    return string;
}

Boolean GnDocument ::
write_to_file( int fd )
{
    ostream ostr(fd);
    
    if( ! write_header_to_stream( ostr ) )
	return False;
    if( ostr.fail() )
	return False;

    if( ! write_to_stream( ostr ) )
	return False;
    if( ostr.fail() )
	return False;

    if( SaveHistory() && ! WriteHistory(ostr) )
	return False;
    if( ostr.fail() )
	return False;
	
    return True;
}

Boolean GnDocument ::
read_from_file( int fd, Boolean create_win )
{
    istream istr(fd);

    Dimension new_width, new_height;

    if( ! read_header_from_stream(istr, &new_width, &new_height) )
	return False;
    if( istr.fail() )
	return False;
    
    if( create_win ) {
	create_windows(new_width, new_height);
	rename(file_pathname);
    } else {
	resize_windows(new_width, new_height);
    }

    if( ! read_from_stream( istr ) )
	return(False);
    if( istr.fail() )
	return False;

    if( SaveHistory() && ! RestoreHistory(istr) )
	return(False);
    if( istr.fail() )
	return False;
	
    modified = FALSE;
    if( save_menu_entry && ! save_menu_entry->get_sensitive() )
	save_menu_entry->set_sensitive(True);
    command_executed_update();
    if( history_scroller )
        history_scroller->update();

    return True;
}

Boolean GnDocument ::
write_header_to_stream( ostream &ostr )
{
    put_string(ostr, GnApplication::get()->app_signature(), False );
    ostr << main_shell->getR_width() << " ";
    ostr << main_shell->getR_height() << "\n";
    return ! ostr.fail();
}

Boolean GnDocument ::
read_header_from_stream(istream &istr, Dimension *width, Dimension *height)
{
    Dimension w, h;
    char *signature;
    Boolean valid;

    signature = get_string(istr);
    
    if(istr.fail() || signature == 0 ||
       (istr >> w, istr.fail()) || (istr >> h, istr.fail()))
	return(False);
    *width = w;
    *height = h;

    valid = strcmp(signature, GnApplication::get()->app_signature());
    delete signature;
    
    return valid == 0 ? True : False;
}

Boolean GnDocument::
write_to_stream( ostream & )
{
    return(True);
}

Boolean GnDocument::
read_from_stream( istream & )
{
    return(True);
}

Boolean GnDocument::
SaveHistory()
{
    return False;
}

GnCommand *GnDocument::
MakeNewCommand(char *)
{
    return 0;
}

Boolean GnDocument::
WriteHistory(ostream &ostr)
{
    REQUIRE( SaveHistory(), "Document shall save command history");
    
    int n_undos = 0;
    int n_redos = 0;

    for(GnCommandListIterator ut(undo_commands);
	! ut.Off() && ut.Item()->IsStorable(); n_undos++, ut.Forth() );

    for(GnCommandListIterator rt(redo_commands);
	! rt.Off() && rt.Item()->IsStorable(); n_redos++, rt.Forth());

    //cout << "n_undos = " << n_undos << " n_redos = " << n_redos << "\n";
    //cout.flush();
    
    ostr << n_undos << " " << n_redos << "\n";
    if( ostr.fail() )
	return False;

    if( ut.Off() )
	ut.Start();
    else
	ut.Forth();
    
    for( ; ! ut.Off(); ut.Forth() ) {
	put_string(ostr, ut.Item()->Class()->Name());
	if( ostr.fail() )
	    return False;
	if( ! ut.Item()->WriteToStream(ostr) )
	    return False;
    }

    for(rt.Start(); ! rt.Off() && rt.Item()->IsStorable(); rt.Forth()) {
	put_string(ostr, rt.Item()->Class()->Name());
	if( ostr.fail() )
	    return False;	
	if( ! rt.Item()->WriteToStream(ostr) )
	    return False;
    }
    return True;
}

Boolean GnDocument::
RestoreHistory(istream &istr)
{
    REQUIRE( SaveHistory(), "Document shall save command history");
    
    int n_undos;
    int n_redos;

    istr >> n_undos;
    istr >> n_redos;
    
    //cout << "n_undos = " << n_undos << " n_redos = " << n_redos << "\n";
    //cout.flush();

    for( int i = 0; i < n_undos; i++ ) {
	char *command_class_name = get_string(istr);
	if( istr.fail() || command_class_name == 0 )
	    return False;
	//cout << "Command Class = " << command_class_name << "\n";
	//cout.flush();
	GnCommand *new_command = MakeNewCommand(command_class_name);
	delete command_class_name;
	if( new_command == 0 ) {
	    cout << "Failed To allocate command object !!!\n";
	    cout.flush();
	    return False;
	}
	if( new_command->ReadFromStream(istr) == False || istr.fail() )
	    return False;
	undo_commands.Append(new_command);
    }

    for( i = 0; i < n_redos; i++ ) {
	char *command_class_name = get_string(istr);
	if( istr.fail() || command_class_name == 0 )
	    return False;
	//cout << "Command Class = " << command_class_name << "\n";
	//cout.flush();
	GnCommand *new_command = MakeNewCommand(command_class_name);
	delete command_class_name;
	if( new_command == 0 ) {
	    cout << "Failed To allocate command object !!!\n";
	    cout.flush();
	    return False;
	}
	if( new_command->ReadFromStream(istr) == False || istr.fail() )
	    return False;
	redo_commands.Append(new_command);
    }
    
    return True;
}

//void GnDocument ::
//create_checkpoint ( iostream *os )
//{
    //write_to_stream( *os );
//}

//void GnDocument ::
//restore_checkpoint ( iostream *is )
//{
    //read_from_stream( *is );
//}

void GnDocument ::
postscript_print () {}

void GnDocument ::
bitmap_print () {}

void GnDocument ::
display_history_scroller ()
{
    if ( history_scroller == 0 ) {
	history_scroller = new GnHistoryScroller( this );
	history_scroller->create( main_shell, "HistoryScroller" );
    }
    history_scroller->manage();
}

void GnDocument ::
receive_activate_callback(caddr_t /* call_data */)
{
    GnApplication::get()->get_application_shell()->beep();
}

void GnDocument ::
receive_minifinder_callback(caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_minifinder_callback");
    minifinder ();
}

void GnDocument ::
receive_do_close_callback(caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_do_close_callback");
    if( do_close () == OK ) {
        GnApplication::get()->destroy_document(this);
    }
}

void GnDocument ::
receive_save_as_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_save_as_callback");
    save_as ();
}

void GnDocument ::
receive_save_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_save_callback");
    save ();
}

void GnDocument ::
receive_revert_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_revert_callback");
    revert ();
}

void GnDocument ::
receive_postscript_print_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_postscript_print_callback");
    postscript_print ();
}

void GnDocument ::
receive_bitmap_print_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_bitmap_print_callback");
    bitmap_print ();
}

void GnDocument ::
receive_undo_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_undo_callback");
    undo ();
}

void GnDocument ::
receive_redo_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_redo_callback");
    redo ();
}

void GnDocument ::
receive_replay_history_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_replay_history_callback");
    replay_history ();
}

void GnDocument ::
receive_display_history_scroller_callback( caddr_t /*call_data*/ )
{
    Gina_Debug_NL ("GnDocument :: receive_display_history_scroller_callback");
    display_history_scroller ();
}

