// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnGraphView.C   Version 1.4
//   LastSCCS    : 2/27/92  10:47:35
//   LastEdit    : "Thu Feb 27 09:05:24 1992"
//   Description : 
//   Author      : Andreas Genau
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_2(GnGraphView,GnMultipleSelection,GnView,);

GnGraphView::
GnGraphView(GnDocument *doc, unsigned int number_of_chars_in_node)
: GnView(doc)
{
    m_node_width  = 0;
    m_node_height = 0;
    m_distance = 10;
    m_vertical_distance = m_distance;
    m_horizontal_distance = m_distance;
    m_node_number_of_chars = number_of_chars_in_node;
    m_direction = eVertical;
    m_layouter = new GnLayout();
    m_layouter->SetNodeList(&entries);
    m_layouter->SetEdgeList(&m_edges);
}

void GnGraphView::
StringDraw(XmString string, int x, int y, Dimension width)
{
    XmStringDraw(document->get_display(), x_window, fontlist, string,
		 default_text_gc, x, y, width, XmALIGNMENT_BEGINNING,
		 XmSTRING_DIRECTION_L_TO_R, NULL);
}

void GnGraphView::
StringExtent(XmString string, Dimension *width, Dimension *height)
{
    XmStringExtent(fontlist, string, width, height);
}

GnViewNode *GnGraphView::
FindViewNode(GnPresentable *node)
{
    Require(node);
    return(GnViewNode::TypeSafeDownCast(FindSelectionItem(node)));
}

void GnGraphView::
AddEdge(GnPresentable *source, GnPresentable *destination)
{
    Require(source); Require(destination);
    GnViewNode *source_item, *dest_item;
    if ((source_item = FindViewNode(source))
	    &&(dest_item = FindViewNode(destination))){
	GnViewEdge *new_edge = new GnViewEdge(this, source_item, dest_item);
	source_item->AddEdgeTo(new_edge);
	dest_item->AddEdgeFrom(new_edge);
	m_edges.Append(new_edge);
	new_edge->Visualize();
    }   }

void GnGraphView::
SetVerticalDistance(Dimension d)
{
    Require(m_layouter);
    m_vertical_distance = d;
    m_layouter->SetVerticalDistance(d);
    DoLayout();
}

void GnGraphView::
SetHorizontalDistance(Dimension d)
{
    Require(m_layouter);
    m_horizontal_distance = d;
    m_layouter->SetHorizontalDistance(d);
    DoLayout();
}

GnLayout *GnGraphView::
SelectLayout(GnLayout *layouter)
{
    Require(layouter);
    GnLayout *old_layouter = m_layouter;
    m_layouter = layouter;
    m_layouter->SetNodeList(&entries);
    m_layouter->SetEdgeList(&m_edges);
    m_layouter->SetVerticalDistance(m_vertical_distance);
    m_layouter->SetHorizontalDistance(m_horizontal_distance);
    DoLayout();
    return old_layouter;
}

void GnGraphView::
DeleteGraph()
{
    while (!entries.Empty())
	Remove(GnViewNode::TypeSafeDownCast(entries.First())->GetValue());
    force_redraw();
}

void GnGraphView::
DeleteEdge(GnViewEdge *edge)
{
    LIST_ITERATE(&m_edges,GnViewEdgeListIterator,it)
	if (it.Item() == edge){
	    it.Remove();
	    delete edge;
            return;
	}
}

void GnGraphView::
CallSelectionCallback()
{
    if (call_callbacks)
	call_selection_callback();
}

void GnGraphView::
DoLayout()
{
    Require(m_layouter);
    m_layouter->DoNodeLayout();
    resized();
    m_layouter->DoEdgeLayout();
    force_redraw();
}

void GnGraphView::
resized()
{
    int upper_border = 0, lower_border = 0, left_border = 0, right_border = 0;
    Dimension node_width = 0, node_height = 0;
    GnViewNode *item;
    LIST_ITERATE(&entries,GnSelectionItemListIterator,it){
	item = GnViewNode::TypeSafeDownCast(it.Item());
	upper_border  = MIN(upper_border,item->get_y_pos());
	lower_border  = MAX(lower_border,item->get_y_pos());
	left_border   = MIN(left_border,item->get_x_pos());
	right_border  = MAX(right_border,item->get_x_pos());
	node_width  = MAX(node_width,item->get_width());
	node_height = MAX(node_height,item->get_height());
    }
    if (left_border != 0 || upper_border != 0)
	LIST_ITERATE_WITH(it){
	    item = GnViewNode::TypeSafeDownCast(it.Item());
	    item->move(item->get_x_pos() - left_border,
		       item->get_y_pos() - upper_border, False);
	}
    right_border  = right_border + node_width + 1;
    lower_border  = lower_border + node_height + 1;
    setR_height(MIN(MAX(lower_border - upper_border,100),65565));
    setR_width(MIN(MAX(right_border - left_border,100),65565));
    SetValues();
}

GnSelectionItem *GnGraphView::
create_item(GnPresentable *obj, Boolean sel, Boolean sens)
{
    return(new GnViewNode(this, obj, m_node_number_of_chars));
}

void GnGraphView::
Removed(GnSelectionItem *p_item)
{
    (GnViewNode::TypeSafeDownCast(p_item))->ClearReferences(); 
}

void GnGraphView::
VisualizeItem(GnSelectionItem *item)
{
    Require(IsMotifWidgetCreated());
    (GnViewNode::TypeSafeDownCast(item))->Visualize();
}

void GnGraphView::
UnvisualizeItem(GnSelectionItem *item)
{
    (GnViewNode::TypeSafeDownCast(item))->Hide();
}


