// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnView.C   Version 1.46
//   LastSCCS    : 2/25/92  12:35:30
//   LastEdit    : "Mon Jan 20 14:11:57 1992"
//   Description : 
//   Authors     : Robert Babatz and Andreas Baecker
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnView,GnDrawingArea,);

ImplGenericListWithIterator(GnViewList,GnViewListIterator);

GnView ::
GnView(GnDocument *doc)
{
    document = doc;
    backing_store = 0;
    x_window = 0;
    drawing_is_possible = False;
    gcontext = 0;
    printing = 0;
    propagate_button_press = TRUE;
    repeated_click = 0;
    last_click_time = 0;
    double_click_intervall = 500;
    Overlapping_objects = TRUE;
    something_selected = False;

    font_name = 0;
    font = 0;
    font_info = 0;
    fontlist = 0;

    // Overridden later with true back- and foreground !
    foreground = 0;
    background = 1;
    
}

Boolean GnView::DrawingIsPossible()
{
    return  drawing_is_possible;
}

Widget GnView::create_widget(GnWidget *parent, char *w_name)
{
    return(GnDrawingArea::create_widget(parent, w_name));
}

void GnView ::
before_create (GnWidget *)
{
//    setR_shadowThickness(6);
}

void GnView ::
after_create ()
{
    Parent()->register_scrolled_child( this );
    add_resizeCallback(CALLBACK(GnView, receive_resize_callback, this));
    add_event_handler(CALLBACK(GnView, handle_expose, this), ExposureMask);
    add_event_handler(CALLBACK(GnView, handle_button_press, this),
		      ButtonPressMask);
    add_event_handler(CALLBACK(GnApplication, receive_button_release_event,
			       APPLICATION), ButtonReleaseMask);
    add_event_handler(CALLBACK(GnApplication, receive_button_1_motion_event,
			       APPLICATION), Button1MotionMask);
    add_event_handler(CALLBACK(GnView, handle_key_press, this), KeyPressMask);
    add_event_handler(CALLBACK(GnView, handle_key_release,this),KeyReleaseMask);
    // Auskommentiert, da enclosing-shell auch NULL liefern kann (AB)
    // par->enclosing_shell()->get_views().append( this );

    // Get font attributes

    font_name = "fixed";
    font = XLoadFont(get_display(), font_name);
    font_info = XQueryFont(get_display(), font);
    fontlist = XmFontListCreate(font_info, XmSTRING_DEFAULT_CHARSET);

    Ensure(font_info != 0);

}

void GnView ::
receive_resize_callback(caddr_t /* call_data */ )
{
    resized();
}

void GnView::
determine_window_attributes()
{
    Require(widget_id != 0);

    x_window = XtWindow( widget_id );
    foreground = getR_foreground();
    background = getR_background();
    
    XGCValues val;
    val.function = GXcopy;
    val.foreground = foreground;
    val.background = background;
    val.line_width = 0;
    val.font = font;

    // Get font attributes

//    font_name = "fixed";
//    val.font = font = XLoadFont(get_display(), font_name);
//    font_info = XQueryFont(get_display(), font);
//    fontlist = XmFontListCreate(font_info, XmSTRING_DEFAULT_CHARSET);

//    Ensure(font_info != 0);

    gcontext = XCreateGC(get_display(), x_window,
			 GCFunction | GCForeground | GCBackground | GCFont,
			 &val);
    text_gc = XCreateGC(get_display(), x_window,
			GCFunction | GCForeground | GCBackground | GCFont,
			&val);
    default_text_gc = XCreateGC(get_display(), x_window,
				GCFunction | GCForeground|GCBackground|GCFont,
				&val);
    
    val.function   = GXxor;
    val.line_style = LineOnOffDash;
    feedback_gc = XCreateGC(get_display(), x_window,
			    GCForeground | GCBackground | GCFunction |
			    GCLineStyle | GCFont,
			    &val );
     drawing_is_possible = True;
}

void GnView ::
handle_expose(caddr_t call_data)
{
    XExposeEvent *event = (XExposeEvent *)call_data;

    Gina_Debug("GnView :: handle_expose c=");
    Gina_Debug(event->count);Gina_Debug(" x=");
    Gina_Debug(event->x);Gina_Debug(" y=");
    Gina_Debug(event->y);Gina_Debug(" w=");
    Gina_Debug(event->width);Gina_Debug(" h=");
    Gina_Debug_NL(event->height);
    if ( ! x_window )
	determine_window_attributes();

    GnMouseDownCommand *cmd = APPLICATION->get_active_mouse_command();

//     XRectangle clip_rect;
//     clip_rect.x = 0;
//     clip_rect.y = 0;
//     clip_rect.width = event->width;
//     clip_rect.height = event->height;
//     XSetClipRectangles(get_display(), gcontext, event->x, event->y,
// 		       &clip_rect, 1, Unsorted);

    if ( cmd ) {
	cmd->draw_feedback( cmd->get_last_x(), cmd->get_last_y(), TRUE );
	// Kann m.E. weg (AB)
	//XClearArea( get_display(), x_window,
		    //event->x, event->y, event->width, event->height, FALSE );
    }
    execute_draw( event->count, event->x, event->y, event->width,
		  event->height );
    if ( cmd )
	cmd->draw_feedback( cmd->get_last_x(), cmd->get_last_y(), FALSE );

//    XSetClipMask(get_display(), gcontext, None);
}

void GnView ::
handle_button_press(caddr_t call_data)
{
    XButtonEvent *event = (XButtonEvent *)call_data;

    Gina_Debug("GnView :: handle_button_press ");
    Gina_Debug(event->state);Gina_Debug(" ");
    Gina_Debug(event->button);Gina_Debug(" x=");
    Gina_Debug(event->x);Gina_Debug(" y=");
    Gina_Debug_NL(event->y);

    unsigned int modifiers = event->state;
    unsigned int code = event->button;
    Time timestamp = event->time;

    if(! APPLICATION->get_active_mouse_command() || modifiers >= 256) {

	// Compute virtual code
	BUTTON_CODE virtual_code = OTHER;
	if ( modifiers <= 5 ) {
	    switch ( code*10 + modifiers ) {
	      case 10: // Button1
		virtual_code = SELECT;
		break;
	      case 11: // Button1 + Shift
		virtual_code = EXTEND;
		break;
	      case 14: // Button1 + Ctrl
		virtual_code = TOGGLE;
		break;
	      case 20: // Button2
		virtual_code = INSERT;
		break;
	      case 30: // Button3
		virtual_code = MENU;
		break;
	    }
	}

	// Check for repeated clicks
	if ( last_click_time &&
	     (timestamp - last_click_time) < double_click_intervall )
	    repeated_click++;
	else
	    repeated_click = 1;
	last_click_time = timestamp;

	button_press_before(virtual_code, repeated_click, event->x, event->y);
    }
}

void GnView ::
button_press_before( int code, int repetition, int x, int y )
{
    Gina_Debug( "GnView :: button_press_before " );
    Gina_Debug(code);Gina_Debug(" ");Gina_Debug(repetition);Gina_Debug(" x=");
    Gina_Debug(x);Gina_Debug(" y=");Gina_Debug_NL(y);
    
    // Propagate button-press event to view-objects
    Boolean hit = FALSE;
    if ( propagate_button_press && (code != INSERT) ) {
	GnViewObject *obj;
	GnViewObjectIterator iterator(&view_objects);
	for( iterator.Start(); !iterator.Off(); iterator.Forth() ) {
	    obj = iterator.Item();
	    if((obj->is_mouse_sensitive()) &&
	       (obj->point_inside( x, y )) ) {
		obj->button_press(code, repetition,
				  x - obj->get_x_pos(), y - obj->get_y_pos() );
		hit = TRUE;
		break;
	    }
	}
    }
    // Call view's button-press or insert_selection method
    //   if event was not propagated
    if( ! propagate_button_press || ! hit )
	if ( code == INSERT )
	    insert_selection( x, y );
	else
	    button_press(code, repetition, x, y);
}

void GnView ::
button_press( int code, int repetition, int x, int y )
{
    Gina_Debug( "GnView :: button_press " );
    Gina_Debug(code);Gina_Debug(" ");Gina_Debug(repetition);Gina_Debug(" x=");
    Gina_Debug(x);Gina_Debug(" y=");Gina_Debug_NL(y);

    if ( code == SELECT )
	(new GnObjectSelector( this, x, y ))->submit();
}

void GnView ::
key_press ( char, int, unsigned int, int, int )
{
    beep ();
}

void GnView ::
key_release( char, int, unsigned int, int, int )
{
    beep();
}

void GnView ::
handle_key_press(caddr_t call_data)
{
    XKeyEvent *event = (XKeyEvent *)call_data;

    key_press( ' ', event->keycode, event->state, event->x, event->y );
}

void GnView ::
handle_key_release (caddr_t call_data)
{
    XKeyEvent *event = (XKeyEvent *)call_data;

    key_release( ' ', event->keycode, event->state, event->x, event->y );
}

// XtSelectionCallbackProc:
void
receive_selection(Widget /* w */, XtPointer client_data, Atom* selection,
		  Atom* type, XtPointer value,
		  unsigned long* length, int* format)
{
    Gina_Debug("GnView :: receive_selection ");
    Gina_Debug(((int*)client_data)[0]);Gina_Debug(" ");
    Gina_Debug(((int*)client_data)[1]);Gina_Debug(" ");
    Gina_Debug(*selection);Gina_Debug(" ");
    Gina_Debug(*type);Gina_Debug(" ");
    Gina_Debug((char*)value);Gina_Debug(" ");
    Gina_Debug(*length);Gina_Debug(" ");
    Gina_Debug_NL(*format);
}

void GnView ::
insert_selection( int x, int y )
{
    Gina_Debug("GnView :: insert_selection ");
    Gina_Debug(x);Gina_Debug(" ");
    Gina_Debug(y);Gina_Debug(" ");
    Gina_Debug(XA_PRIMARY);Gina_Debug(" ");
    Gina_Debug_NL(XA_STRING);

    int *ab = new int [2];
    ab[0] = x;
    ab[1] = y;
    XtGetSelectionValue(widget_id, XA_PRIMARY, XA_STRING,
			receive_selection, (XtPointer)ab, last_click_time );
}

Boolean GnView ::
unselect_all_objects()
{
    Boolean unselected = False;

    if ( something_selected ) {
	GnViewObjectIterator iterator(view_objects);
	GnViewObject *object;

	for ( iterator.Start(); ! iterator.Off(); iterator.Forth() ) {
	    object = iterator.Item();
	    if ( object->facilities() &
		(OBJ_SELECTABLE|OBJ_MOVABLE|OBJ_RESIZABLE) )
		if( object->is_selected()) {
		    object->unselect(False);
		    unselected = True;
		}
	}
	something_selected = False;
	XtDisownSelection(widget_id, XA_PRIMARY, last_click_time);
    }
    return unselected;
}

// XtConvertSelectionProc:
Boolean
convert_selection(Widget /* w */, Atom* selection, Atom* target, Atom* type,
		  XtPointer* value, unsigned long* length, int* format)
{
    Gina_Debug("GnView :: convert_selection ");

    *type = *target;
    char *txt = strdup("Test Text");
    *value = txt;
    *length = 10;
    *format = 8;

    Gina_Debug(*selection);Gina_Debug(" ");
    Gina_Debug(*target);Gina_Debug(" ");
    Gina_Debug(*type);Gina_Debug(" ");
    Gina_Debug(txt);Gina_Debug(" ");
    Gina_Debug(*length);Gina_Debug(" ");
    Gina_Debug_NL(*format);

    return True;
}

// XtLoseSelectionProc:
void
lose_selection(Widget /* w */, Atom* selection)
{
    Gina_Debug("GnView :: lose_selection ");
    Gina_Debug_NL(*selection);

//    unselect_all_objects(); geht erst, wenn Callback funktioniert
}

void GnView ::
register_selection( GnViewObject *, Boolean register_obj )
{
// Der View sollte sich merken, welche Objekte select-iert sind, dann
// kann er pruefen, ob noch was select-iert ist oder nicht, d.h.
// ob ein Objekt das letzte ist, das unselect-iert wird.
    if ( register_obj ) {
	if ( ! something_selected )
	    something_selected =
		XtOwnSelection(widget_id, XA_PRIMARY, last_click_time,
			       convert_selection, lose_selection, NULL);
    }
    else { // unregister selection
	something_selected = False; // to be replaced
	XtDisownSelection(widget_id, XA_PRIMARY, last_click_time);
    }
}

void GnView ::
clear_area ( int x, int y, Dimension wid, Dimension hei )
{
    XClearArea( get_display(), x_window, x, y, wid, hei, TRUE );
}

void GnView ::
force_redraw ()
{
    Gina_Debug_NL("FORCE REDRAW !!!!");
    if ( x_window ) {
	XClearArea( get_display(), x_window, 0, 0, 0, 0, FALSE );
	execute_draw( 0, 0, 0, getR_width(), getR_height() );
    }
}

void GnView ::
invalidate_rectangle ( int x, int y, Dimension wid, Dimension hei )
{
    if ( x_window ) {
	Gina_Debug("GnView :: invalidate_rectangle ");
	Gina_Debug(x);Gina_Debug(" ");Gina_Debug(y);Gina_Debug(" ");
	Gina_Debug(wid);Gina_Debug(" ");Gina_Debug_NL(hei);
// send_expose_event = TRUE oder FALSE: ???
	XClearArea( get_display(), x_window, x, y, wid+1, hei+1, FALSE );
// kann raus, falls send_expose_event=TRUE in XClearArea: !?
	execute_draw( 0, x, y, wid+1, hei+1 );
    }
}

void GnView ::
execute_draw ( int count, int x, int y, Dimension wid, Dimension hei )
{
    draw( count, x, y, wid, hei );
    after_draw( count, x, y, wid, hei );
}

void GnView ::
after_draw ( int /* count */, int x, int y, Dimension wid, Dimension hei )
{
    Gina_Debug( "GnView :: after_draw #obj=");
    Gina_Debug( view_objects.Length() );
    Gina_Debug(" RECT=(");Gina_Debug(x);Gina_Debug(",");Gina_Debug(y);
    Gina_Debug(") (");Gina_Debug(wid);Gina_Debug(",");Gina_Debug(hei);
    Gina_Debug_NL(")");

    GnViewObject *view_obj;

    GnViewObjectIterator iterator(&view_objects);

    for(iterator.Start(); ! iterator.Off(); iterator.Forth() ) {
	view_obj = iterator.Item();
	if( view_obj->inside_rectangle( x, y, wid, hei ) )
	    view_obj->execute_draw( 0, 0, 0, view_obj->get_width(),
				   view_obj->get_height() );
    }
}

void GnView ::
draw ( int count, int x, int y, Dimension wid, Dimension hei ) {}

void GnView ::
resized () {}

void GnView ::
draw_into_pixmap () {}

void GnView ::
change_gc( XGCValues *val, unsigned long value_mask )
{
    XChangeGC( get_display(), gcontext, value_mask, val );
}

void GnView ::
draw_point ( int x, int y )
{
    XDrawPoint( get_display(), x_window, gcontext, x, y );
}

void GnView ::
draw_points ( XPoint *points, int npoints, Boolean relative )
{
    XDrawPoints( get_display(), x_window, gcontext, points, npoints,
		 relative ? CoordModePrevious : CoordModeOrigin );
}

void GnView ::
draw_line ( int x1, int y1, int x2, int y2, Boolean relative )
{
    if ( relative ) {
	x2 += x1;
	y2 += y1;
    }
    XDrawLine( get_display(), x_window, gcontext, x1, y1, x2, y2 );
}

void GnView ::
draw_lines ( XPoint *points, int npoints, Boolean rel, Boolean fill, int shape)
{
    Require( shape == Complex || shape == Convex || shape == Nonconvex );
    XDrawLines( get_display(), x_window, gcontext, points, npoints,
	        rel ? CoordModePrevious : CoordModeOrigin );
    if ( fill )
	XFillPolygon( get_display(), x_window, gcontext, points, npoints, shape,
		      rel ? CoordModePrevious : CoordModeOrigin );
}

void GnView ::
draw_segments ( XSegment *segments, int nsegments )
{
    XDrawSegments( get_display(), x_window, gcontext,
		   segments, nsegments );
}

void GnView ::
draw_rectangle ( int x, int y, Dimension wid, Dimension hei, Boolean fill )
{
    XDrawRectangle( get_display(), x_window, gcontext, x, y, wid, hei );
    if ( fill )
	XFillRectangle( get_display(), x_window, gcontext, x, y, wid, hei );
}

void GnView ::
draw_rectangles ( XRectangle *rects, int nrects, Boolean fill )
{
    XDrawRectangles( get_display(), x_window, gcontext, rects, nrects );
    if ( fill )
	XFillRectangles( get_display(), x_window, gcontext, rects, nrects );
}

void GnView ::
draw_arc ( int x, int y, Dimension wid, Dimension hei,
	   int ang1, int ang2, Boolean fill )
{
    XDrawArc( get_display(), x_window, gcontext, x, y, wid, hei, ang1, ang2 );
    if ( fill )
	XFillArc(get_display(), x_window, gcontext, x, y, wid, hei, ang1, ang2);
}

void GnView ::
draw_arcs ( XArc *arcs, int narcs, Boolean fill )
{
    XDrawArcs( get_display(), x_window, gcontext, arcs, narcs );
    if ( fill )
	XFillArcs( get_display(), x_window, gcontext, arcs, narcs );
}

void GnView ::
draw_glyph ( int x, int y, char c )
{
    XDrawString( get_display(), x_window, gcontext, x, y, &c, 1 );
}

void GnView ::
draw_glyphs ( int x, int y, char* str )
{
    XDrawString( get_display(), x_window, gcontext, x, y, str, strlen(str) );
}

void GnView ::
execute_expose_callback( void * /*client_data*/, caddr_t /*call_data*/ )
{
    execute_draw( 0, 0, 0, 0, 0 );
}

void GnView ::
execute_input_callback( void * /*client_data*/, caddr_t /*call_data*/ )
{
    execute_draw( 0, 0, 0, 0, 0 );
}

void GnView::
setGC_line_width( int width )
{
    Require( width >= 0 );

    XGCValues values;
    values.line_width = width;
    change_gc(&values, GCLineWidth);
}

void GnView ::
DrawGlyphs(GC gc, int x, int y, char* str, unsigned int length)
{
    XDrawString( get_display(), x_window, gc, x, y, str, length );
}

void GnView::
DrawStringAligned(GC gc, char *string, unsigned int length,
		  int x, int y, Dimension width, Dimension height,
		  unsigned char alignment, Dimension offset)
{
    Require(font_info != 0);
    Require(string != 0);
    //Require(length > 0);
    Require(alignment == XmALIGNMENT_CENTER ||
	    alignment == XmALIGNMENT_BEGINNING ||
	    alignment == XmALIGNMENT_END);
    
    XCharStruct sm;
    int direction, font_ascent, font_descent;

    XTextExtents(font_info, string, length, &direction,
		 &font_ascent, &font_descent, &sm);

//     cout << direction << " " << font_ascent << " " << font_descent << " ";
//     cout << sm.lbearing << " ";
//     cout << sm.rbearing << " ";
//     cout << sm.width << " ";
//     cout << sm.ascent << " ";
//     cout << sm.descent << " ";
//     cout << "\n";

    switch( alignment ) {
      case XmALIGNMENT_CENTER:
	DrawGlyphs(gc, x + (width - (sm.rbearing - sm.lbearing))/2,
		   y + (height - (font_ascent+font_descent))/2 + font_ascent,
		   string, length);
	break;
      case XmALIGNMENT_BEGINNING:
	DrawGlyphs(gc, x + offset,
		   y + (height - (font_ascent+font_descent))/2 + font_ascent,
		   string, length);
	break;
      case XmALIGNMENT_END:
	DrawGlyphs(x + width - (sm.rbearing - sm.lbearing) - offset,
		   y + (height - (font_ascent+font_descent))/2 + font_ascent,
		   string, length);
	break;
      default:
	break;
    }
}

Boolean GnView::
IntersectRectangles(int x1, int y1, Dimension w1, Dimension h1,
		    int x2, int y2, Dimension w2, Dimension h2)
{
    return (x2 >= x1 && x2 < x1 + w1 &&
	    ((y2 >= y1 && y2 < y1 + h1) || (y1 >= y2 && y1 < y2 + h2 ))) ||
	   (x1 >= x2 && x1 < x2 + w2 &&
	    ((y2 >= y1 && y2 < y1 + h1) || (y1 >= y2 && y1 < y2 + h2 )));
}

Boolean GnView::
PointInRectangle(int x1, int y1,
		 int x2, int y2, Dimension w2, Dimension h2)
{
    return x1 >= x2 && x1 < x2 + w2 && y1 >= y2 && y1 < y2 + h2;
}

