// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnViewNode.C   Version 1.4
//   LastSCCS    : 2/27/92  10:47:39
//   LastEdit    : "Thu Feb 27 09:05:27 1992"
//   Description : 
//   Author      : Andreas Genau
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_2(GnViewNode,GnSelectionItem,GnDirectManipulationObject,);

ImplGenericListWithIterator(GnViewNodeList,GnViewNodeListIterator);

GnViewNode::
GnViewNode(GnGraphView *view, GnPresentable *node, unsigned int length)
:GnSelectionItem(node, FALSE, TRUE),
 GnDirectManipulationObject(view->get_document()->gc_values(), 10, 10, True)
{
    m_view = view;
    m_hidden = True;
    m_visited = 0;
    Facilities = OBJ_SELECTABLE;
    if (node->IsStringPresentable()){
	char *buffer = new char[length];
	m_label = GnXMSTRING_LTOR(node->PrintOn(buffer, length));
	m_view->StringExtent(m_label, &width, &height);
    }
    m_x_string_distance_to_frame = 2;
    m_y_string_distance_to_frame = 2;
    width += 2 * m_x_string_distance_to_frame;
    height += 2 * m_y_string_distance_to_frame;
}

void GnViewNode::
WhereEdgeMayBegin(int angle, int *x, int *y)
{
    switch (angle){
      case 0:
	*x = x_pos + width;
	*y = y_pos + height/2;
	break;
      case 90:
	*x = x_pos + width/2;
	*y = y_pos;
	break;
      case 180:
	*x = x_pos;
	*y = y_pos + height/2;
	break;
      case 270:
	*x = x_pos + width/2;
	*y = y_pos + height;
	break;
      default:
	*x = x_pos + width/2;
	*y = y_pos + height/2;
	break;
    }
}

void GnViewNode::
WhereEdgeMayEnd(int angle, int *x, int *y)
{
    WhereEdgeMayBegin(angle, x, y);
}

void GnViewNode::
WhereEdgeMayBegin(GnViewNode *destination, int *x, int *y)
{
    *x = x_pos + width; *y = y_pos + height/2;
}

void GnViewNode::
WhereEdgeMayEnd(GnViewNode *source, int *x, int *y)
{
    *x = x_pos; *y = y_pos + height/2;
}

void GnViewNode::
AddEdgeTo(GnViewEdge *edge)
{
    m_edges_to.Append(edge);
}

void GnViewNode::
AddEdgeFrom(GnViewEdge *edge)
{
    m_edges_from.Append(edge);
}

Boolean GnViewNode::
NoEdgeOfMine(GnViewEdge *edge)
{
    for (GnViewEdgeListIterator itt(&m_edges_to); !itt.Off(); itt.Forth())
	if (itt.Item() == edge)
	    return(False);
    for (GnViewEdgeListIterator itf(&m_edges_from); !itf.Off(); itf.Forth())
	if (itf.Item() == edge)
	    return(False);
    return(True);
}

void GnViewNode::
RemoveEdgeTo(GnViewEdge *edge)
{
    m_edges_to.Remove(edge);
}

void GnViewNode::
RemoveEdgeFrom(GnViewEdge *edge)
{
    m_edges_from.Remove(edge);
}

void GnViewNode::
ClearReferences()
{
    GnViewEdge *to_delete;
    while (!m_edges_to.Empty())
       m_view->DeleteEdge(m_edges_to.First());
    while (!m_edges_from.Empty())
       m_view->DeleteEdge(m_edges_from.First());
    Hide();
}

void GnViewNode::
Visualize()
{
    Require(m_view);
    Require(m_view->IsMotifWidgetCreated());
    if (Hidden()){
	install(m_view, 0, 0, False);
	for (GnViewEdgeListIterator itt(&m_edges_to); !itt.Off(); itt.Forth())
	    itt.Item()->Visualize();
	for (GnViewEdgeListIterator itf(&m_edges_from);!itf.Off();itf.Forth())
	    itf.Item()->Visualize();
        m_hidden = False;
}   }

void GnViewNode::
Hide()
{
    if (!Hidden()){
	deinstall(False);
	for (GnViewEdgeListIterator itt(&m_edges_to); !itt.Off(); itt.Forth())
	    itt.Item()->Hide();
	for (GnViewEdgeListIterator itf(&m_edges_from);!itf.Off();itf.Forth())
	    itf.Item()->Hide();
	m_hidden = True;
}   }

void GnViewNode::
became_selected()
{
     Select(True);
     m_view->CallSelectionCallback();
}

void GnViewNode::
became_unselected()
{
     Select(False);
     m_view->CallSelectionCallback();
}

void GnViewNode::
draw(int count, int x, int y, Dimension width, Dimension height)
{
    Require(parent_view);
    if (count == 0){
	if (value->IsStringPresentable())
	    m_view->StringDraw(m_label, x_pos+m_x_string_distance_to_frame+1, 
                                y_pos+m_y_string_distance_to_frame+1, width);
	m_view->draw_rectangle(x_pos, y_pos, width, height);
}   }





