// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnViewObject.C   Version 1.28
//   LastSCCS    : 1/23/92  12:34:08
//   LastEdit    : "Tue Feb 25 10:29:05 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnViewObject,GnObject,);

ImplGenericListWithIterator(GnViewObjectList,GnViewObjectIterator);

GnViewObject ::
GnViewObject( XGCValues& cg_val, Dimension wid, Dimension hei,
	     Boolean mouse_sens)
{
    width = wid;
    height = hei;
    mouse_sensitive = mouse_sens;
    selected = FALSE;
    x_pos = 0;
    y_pos = 0;
    parent_view = 0;
    Facilities = OBJ_MOVABLE | OBJ_SELECTABLE | OBJ_RESIZABLE;
    values = cg_val;
    value_mask = 0;
}

void GnViewObject::
select(Boolean redraw)
{
    if( ! selected ) {
	selected = TRUE;
	parent_view->register_selection( this );
	if( redraw )
	    force_redraw();
        became_selected();
    }
}

void GnViewObject::
unselect(Boolean redraw)
{
    if( selected ) {
	selected = FALSE;
	parent_view->register_selection( this, False );
	if( redraw )
	    force_redraw();
	became_unselected();
    }
}

void GnViewObject::
toggle_selected(Boolean redraw)
{
    selected = ! selected;
    parent_view->register_selection( this, selected );
    if( redraw )
	force_redraw();
    if (selected)
	became_selected();
    else
	became_unselected();
}

void GnViewObject ::
install ( GnView *view, int x, int y, Boolean redraw )
{
    //Gina_Debug_NL( "GnViewObject :: install +x +y" );
    parent_view = view;
    x_pos = x;
    y_pos = y;
    view->view_objects.Append( this );
    if ( redraw )
	force_redraw();
}

void GnViewObject ::
install ( GnView *view, Boolean redraw )
{
    //Gina_Debug_NL( "GnViewObject :: install" );
    parent_view = view;
    view->view_objects.Append( this );
    if ( redraw )
	force_redraw();
}

void GnViewObject ::
deinstall ( Boolean redraw )
{
    Require(parent_view != 0);

    //Gina_Debug_NL( "GnViewObject :: deinstall" );
    parent_view->view_objects.Remove(this);
    if ( redraw )
	if ( parent_view->overlapping_objects() )
	    parent_view->invalidate_rectangle( x_pos, y_pos, width, height );
	else
	    parent_view->clear_area( x_pos, y_pos, width, height );
    parent_view = 0;
}

Boolean GnViewObject ::
point_inside ( int x, int  y )
{
    //Gina_Debug_NL( "GnViewObject :: point_inside" );
    return point_inside_rectangle( x, y, x_pos, y_pos, width, height );
}

Boolean GnViewObject ::
inside_rectangle ( int rect_x, int rect_y,
		   Dimension rect_wid, Dimension rect_hei )
{
    //Gina_Debug( "GnViewObject :: inside_rectangle RECT=(" );
    //Gina_Debug(rect_x);Gina_Debug(",");Gina_Debug(rect_y);
    //Gina_Debug(") (");Gina_Debug((int)rect_wid);
    //Gina_Debug(",");Gina_Debug((int)rect_hei);Gina_Debug(") OBJ=(");
    //Gina_Debug(x_pos);Gina_Debug(",");Gina_Debug(y_pos);
    //Gina_Debug(") (");Gina_Debug((int)width);
    //Gina_Debug(",");Gina_Debug((int)height);Gina_Debug_NL(")");

    return ! ( ( (rect_y + (int)rect_hei) < y_pos ) ||
	       ( rect_y > (y_pos + (int)height)   ) ||
	       ( (rect_x + (int)rect_wid) < x_pos ) ||
	       ( rect_x > (x_pos + (int)width)    ) );
}

void GnViewObject ::
move ( int new_x, int new_y, Boolean redraw )
{
    //Gina_Debug( "GnViewObject :: move " );
    //Gina_Debug(new_x);Gina_Debug(" ");Gina_Debug_NL(new_y);
    int old_x = x_pos;
    int old_y = y_pos;
    x_pos = new_x;
    y_pos = new_y;
    if ( redraw && parent_view && parent_view->x_window )
	if ( parent_view->overlapping_objects () ) {
	    parent_view->invalidate_rectangle( old_x, old_y, width, height );
	    parent_view->invalidate_rectangle( new_x, new_y, width, height );
	}
	else {
	    parent_view->clear_area( old_x, old_y, width, height );
	    execute_draw( 0, 0, 0, width, height );
	}
}

void GnViewObject ::
resize ( Dimension new_width, Dimension new_height, Boolean redraw )
{
    //Gina_Debug_NL( "GnViewObject :: resize" );

    Dimension old_width  = width;
    Dimension old_height = height;

    width  = new_width;
    height = new_height;
    if( redraw && parent_view && parent_view->x_window )
        parent_view->invalidate_rectangle( x_pos, y_pos,
					   MAX( new_width, old_width ),
					   MAX( new_height, old_height ) );
}

void GnViewObject ::
invalidate_rectangle ( int x, int y, Dimension wid, Dimension hei )
{
    if ( parent_view && parent_view->x_window ) {
	//Gina_Debug_NL( "GnViewObject :: invalidate_rectangle " );
	//Gina_Debug(x_pos+x);Gina_Debug(" ");Gina_Debug(y_pos+y);Gina_Debug(" ");
	//Gina_Debug(wid);Gina_Debug(" ");Gina_Debug_NL(hei);
        parent_view->invalidate_rectangle( x_pos + x, y_pos + y, wid, hei );
    }
}

void GnViewObject ::
force_redraw ()
{
    //Gina_Debug_NL( "GnViewObject :: force_redraw" );
    if ( parent_view && parent_view->x_window )
        parent_view->invalidate_rectangle( x_pos, y_pos, width, height );
}

void GnViewObject ::
execute_draw ( int count, int x, int y, Dimension wid, Dimension hei )
{
    draw( count, x, y, wid, hei );
    after_draw( count, x, y, wid, hei );
}

void GnViewObject ::
draw ( int count, int, int, Dimension, Dimension )
{
    if ( count == 0  ) {
	//Gina_Debug( "GnViewObject :: draw " );
	//Gina_Debug(x_pos);Gina_Debug(" ");Gina_Debug(y_pos);Gina_Debug(" ");
	//Gina_Debug(width);Gina_Debug(" ");Gina_Debug_NL(height);
        this->draw_rectangle( x_pos, y_pos, width, height );
    }
}

void GnViewObject ::
after_draw ( int count, int, int, Dimension, Dimension ) {}

void GnViewObject ::
button_press ( int, int, int, int )
{
    parent_view->beep();
}

void GnViewObject::
became_selected()
{}

void GnViewObject::
became_unselected()
{}

void GnViewObject ::
key_press ( int, int, int, int, int )
{
    parent_view->beep();
}

void GnViewObject ::
draw_point ( int x, int y )
{
    //Gina_Debug_NL( "GnViewObject :: draw_point" );
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_point( x_pos + x, y_pos + y );
}

void GnViewObject ::
draw_points( XPoint *points, int npoints, Boolean relative )
{
    REQUIRE( npoints > 0, "At least one point to draw");
	
    XPoint *newp = new XPoint [ npoints ];
    for ( int i = 0; i < npoints; i++ ) {
	newp[i].x = points[i].x + x_pos;
	newp[i].y = points[i].y + y_pos;
    }
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_points( newp, npoints, relative );
    delete newp;
}

void GnViewObject ::
draw_line ( int x1, int y1, int x2, int y2, Boolean rel )
{
    //Gina_Debug_NL( "GnViewObject :: draw_line" );
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_line( x_pos+x1, y_pos+y1, x_pos+x2, y_pos+y2, rel );
}

void GnViewObject ::
draw_lines ( XPoint *points, int npoints, Boolean rel, Boolean fill, int shape )
{
    REQUIRE( npoints > 1, "At least one line to draw");
    REQUIRE(shape==Complex||shape==Convex||shape==Nonconvex,"Valid shape");

    XPoint *newp = new XPoint [ npoints ];
    for ( int i = 0; i < npoints; ++i ) {
	newp[i].x = points[i].x + x_pos;
	newp[i].y = points[i].y + y_pos;
    }
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_lines( newp, npoints, rel, fill, shape );
    delete newp;
}

void GnViewObject ::
draw_segments ( XSegment *segments, int nsegments )
{
    //Gina_Debug_NL( "GnViewObject :: draw_segments" );
    XSegment *news = new XSegment [ nsegments ];
    for ( int i = 0; i < nsegments; ++i ) {
	news[i].x1 = segments[i].x1 + x_pos;
	news[i].y1 = segments[i].y1 + y_pos;
	news[i].x2 = segments[i].x2 + x_pos;
	news[i].y2 = segments[i].y2 + y_pos;
    }
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_segments( news, nsegments );
    delete news;
}

void GnViewObject ::
draw_rectangle ( int x, int y, Dimension wid, Dimension hei, Boolean fill )
{
    //Gina_Debug( "GnViewObject :: draw_rectangle " );
    //Gina_Debug(x_pos+x);Gina_Debug(" ");Gina_Debug(y_pos+y);Gina_Debug(" ");
    //Gina_Debug(wid);Gina_Debug(" ");Gina_Debug_NL(hei);

    register int lw = getGC_line_width()/2;

    parent_view->change_gc( &values, value_mask );
    parent_view->draw_rectangle(x_pos + x + lw, y_pos + y + lw,
				wid - lw - lw, hei - lw - lw, fill );
}

void GnViewObject ::
draw_rectangles ( XRectangle *rectangles, int nrectangles, Boolean fill )
{
    //Gina_Debug_NL( "GnViewObject :: draw_rectangles" );
    XRectangle *newr = new XRectangle [ nrectangles ];
    for ( int i = 0; i < nrectangles; ++i ) {
	newr[i].x = rectangles[i].x + x_pos;
	newr[i].y = rectangles[i].y + y_pos;
	newr[i].width = rectangles[i].width;
	newr[i].height = rectangles[i].height;
    }
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_rectangles( newr, nrectangles, fill);
    delete newr;
}

void GnViewObject ::
draw_arc ( int x, int y, Dimension wid, Dimension hei,
		   int angle1, int angle2, Boolean fill )
{
    //Gina_Debug( "GnViewObject :: draw_arc " );
    //Gina_Debug(x_pos+x);Gina_Debug(" ");Gina_Debug(y_pos+y);Gina_Debug(" ");
    //Gina_Debug(wid);Gina_Debug(" ");Gina_Debug(hei);Gina_Debug(" ");
    //Gina_Debug(angle1);Gina_Debug(" ");Gina_Debug_NL(angle2);

    register int lw = getGC_line_width()/2;

    parent_view->change_gc( &values, value_mask );
    parent_view->draw_arc( x_pos+x+lw, y_pos+y+lw, wid-lw-lw, hei-lw-lw, angle1, angle2, fill );
}

void GnViewObject ::
draw_arcs ( XArc *arcs, int narcs, Boolean fill )
{
    //Gina_Debug_NL( "GnViewObject :: draw_arcs" );
    XArc *newa = new XArc [ narcs ];
    for ( int i = 0; i < narcs; ++i ) {
	newa[i].x = arcs[i].x + x_pos;
	newa[i].y = arcs[i].y + y_pos;
	newa[i].width = arcs[i].width;
	newa[i].height = arcs[i].height;
	newa[i].angle1 = arcs[i].angle1;
	newa[i].angle2 = arcs[i].angle2;
    }
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_arcs( newa, narcs, fill );
    delete newa;
}

void GnViewObject ::
draw_glyph ( int x, int y, char c )
{
    //Gina_Debug_NL( "GnViewObject :: draw_glyph" );
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_glyph( x_pos+x, y_pos+y, c );
}

void GnViewObject ::
draw_glyphs ( int x, int y, char *string )
{
    //Gina_Debug_NL( "GnViewObject :: draw_glyphs" );
    parent_view->change_gc( &values, value_mask );
    parent_view->draw_glyphs( x_pos+x, y_pos+y, string );
}

#define DELIM << " " <<

void GnViewObject ::
write_to_stream( ostream &osf, Boolean FromSubclass )
{
    //Gina_Debug_NL ("GnViewObject :: write_to_stream");
    
    osf << IO_type() DELIM width DELIM height DELIM x_pos DELIM y_pos
	DELIM getGC_line_width();
    
    if( ! FromSubclass )
	osf << endl;
}

void GnViewObject ::
read_from_stream( istream &ifs, Boolean )
{
    //Gina_Debug_NL ("GnViewObject :: read_from_stream");
    int lw;
    ifs >> width >> height >> x_pos >> y_pos >> lw;
    setGC_line_width(lw);
}

// ----------------------------------------------------------------------
// graphics context attributes:
// ----------------------------------------------------------------------
void GnViewObject ::
setGC_function( int p )
{
    Require( p >= GXclear );
    Require( p <= GXset );
    values.function = p;
}

int GnViewObject :: getGC_function()
{
    return values.function;
}

void GnViewObject ::
setGC_plane_mask( unsigned long p )
{
    values.plane_mask = p;
}

void GnViewObject ::
setGC_foreground( unsigned long p )
{
    values.foreground = p;
}

void GnViewObject ::
setGC_background( unsigned long p )
{
    values.background = p;
}

void GnViewObject ::
setGC_line_width( int p )
{
    Require( p >= 0 );
    values.line_width = p;
}

int GnViewObject :: getGC_line_width()
{
    return values.line_width;
}

void GnViewObject ::
setGC_line_style( int p )
{
    Require( p == LineSolid || p == LineDoubleDash || p == LineOnOffDash );
    values.line_style = p;
}

void GnViewObject ::
setGC_cap_style( int p )
{
    Require( p == CapNotLast || p == CapButt || p == CapRound ||
	     p == CapProjecting );
    values.cap_style = p;
}

void GnViewObject ::
setGC_join_style( int p )
{
    Require( p == JoinMiter || p == JoinRound || p == JoinBevel );
    values.join_style = p;
}

void GnViewObject ::
setGC_fill_style( int p )
{
    Require( p == FillSolid || p == FillTiled || p == FillOpaqueStippled ||
	     p == FillStippled );
    values.fill_style = p;
}

void GnViewObject ::
setGC_fill_rule( int p )
{
    values.fill_rule = p;
}

void GnViewObject ::
setGC_arc_mode( int p )
{
    values.arc_mode = p;
}

void GnViewObject ::
setGC_tile( Pixmap p )
{
    values.tile = p;
}

void GnViewObject ::
setGC_stipple( Pixmap p )
{
    values.stipple = p;
}

void GnViewObject ::
setGC_ts_x_origin( int p )
{
    values.ts_x_origin = p;
}

void GnViewObject ::
setGC_ts_y_origin( int p )
{
    values.ts_y_origin = p;
}

void GnViewObject ::
setGC_font( Font p )
{
    values.font = p;
}

void GnViewObject ::
setGC_subwindow_mode( int p )
{
    values.subwindow_mode = p;
}

void GnViewObject ::
setGC_graphics_exposures( Bool p )
{
    values.graphics_exposures = p;
}

void GnViewObject ::
setGC_clip_x_origin( int p )
{
    values.clip_x_origin = p;
}

void GnViewObject ::
setGC_clip_y_origin( int p )
{
    values.clip_y_origin = p;
}

void GnViewObject ::
setGC_clip_mask( Pixmap p )
{
    values.clip_mask = p;
}

void GnViewObject ::
setGC_dash_offset( int p )
{
    values.dash_offset = p;
}

void GnViewObject ::
setGC_dashes( char p )
{
    values.dashes = p;
}
