// This may look like C code, but it is really -*- C++ -*-

//   Module      : GnTimer.C   Version 1.12
//   LastSCCS    : 1/23/92  12:33:20
//   LastEdit    : "Mon Jan 20 13:35:20 1992"
//   Description : 
//   Author      : 
//   Copyright   : GMD Schloss Birlinghoven

META_IMPL_1(GnTimer,GnObject,);

GnTimer::GnTimer()
{
    Require(False);
}

GnTimer::GnTimer(unsigned int timer_interval, GnCallback *cb,
	       Boolean initially_running, Boolean active_in_h,
	       Boolean active_in_loop)
{
    Require(timer_interval > 0);

    destroyed = busy = enabled = False;
    interval = timer_interval;
    running = initially_running;
    active_in_handler = active_in_h;
    active_in_recursive_main_loops = active_in_loop;
    if(cb)
	add_callback(cb);
    if(running)
	install_timeout();
}

GnTimer::~GnTimer()
{
    Require( ! running );
    Require( ! busy );
    
    if(enabled)
	deinstall_timeout();
}

void GnTimer::add_callback(GnCallback *cb)
{
    Require( ! busy );
    Require( cb != 0 );
    
    cb->attach_to_timer(this);
    callbacks.Append(cb);
}

void GnTimer::remove_callback(GnCallback *cb)
{
    Require( ! busy );
    Require( cb != 0 );

    cb->remove();
    callbacks.Remove(cb);
}

void GnTimer::install_timeout()
{
    Require( ! enabled );
    Require( running );

    id = XtAppAddTimeOut(GnWidget::TheToolkit()->AppContext(),
			 interval,
			 (XtTimerCallbackProc)GnCallback::timer_handler,
			 (XtPointer)this);
    enabled = True;
    Ensure( enabled );
}

void GnTimer::deinstall_timeout()
{
    Require( enabled );
    XtRemoveTimeOut(id);
    enabled = False;
    Ensure( ! enabled );
}

void GnTimer::destroy()
{
    Require( ! running );
    Require( ! destroyed );

    destroyed = True;
    if( enabled )
	deinstall_timeout();
}

void GnTimer::start(unsigned int new_timer_interval)
{
    Require( ! running );

    running = True;
    if(new_timer_interval > 0)
	interval = new_timer_interval;
    install_timeout();
}

void GnTimer::restart(unsigned int new_timer_interval)
{
    if( enabled )
	deinstall_timeout();
    running = True;
    if(new_timer_interval > 0)
	interval = new_timer_interval;
    install_timeout();
}

void GnTimer::stop()
{
    Require( running );

    running = False;
    deinstall_timeout();
}



